#! /usr/bin/env python

#Create plot of dem_align results for many input files

import sys
import os
import glob
import numpy as np
import matplotlib.pyplot as plt
from osgeo import gdal
import pandas as pd
from pygeotools.lib import geolib, malib, iolib

#Run as:
#dem_align_post.py $(ll *tif | grep 'alongtrack/WV03' | awk '{print $9}' | sed 's#.tif#_dem_align/*[0-9]_align.tif#')
#dem_align_post.py $(ll *tif | grep 'alongtrack/WV03' | awk '{print $9}' | sed 's#.tif#_dem_align/*tiltcorr_align.tif#')
#dem_align_post.py $(ll *tif | grep 'alongtrack/WV0[12]' | awk '{print $9}' | sed 's#.tif#_dem_align/*tiltcorr_align.tif#')
#dem_align_post.py $(ll *tif | grep 'QB02' | awk '{print $9}' | sed 's#.tif#_dem_align/*align.tif#')
#cat wv3_at_list.txt | sed 's#.tif#_dem_align/*align.tif#' > wv3_at_list_align.txt

outdir = 'dem_align_aster'
#outdir = 'dem_align_at_wv3'
#outdir = 'dem_align_at_wv12'
#outdir = 'dem_align_qb'
#outdir = 'dem_align_noqb'
#outdir = 'dem_align_at'
#outdir = 'dem_align_ct'

if not os.path.exists(outdir):
    os.makedirs(outdir)
out_fn_prefix = os.path.join(outdir, outdir)

#Throw out gross outliers
filter=True
mv_bad=True
#WV/GE
#outlier_mag_thresh = 20 
#ASTER
outlier_mag_thresh = 90

def make_plot3d(x, y, z, title=None, orthogonal_fig=True):
    cmean = np.mean([x,y,z], axis=1)
    cstd = np.std([x,y,z], axis=1)
    cmed = np.median([x,y,z], axis=1)
    cnmad = malib.mad([x,y,z], axis=1)
    x_corr = x - cmean[0]
    y_corr = y - cmean[1]
    z_corr = z - cmean[2]
   
    ce90 = geolib.CE90(x,y)
    ce90_corr = geolib.CE90(x_corr,y_corr)
    le90 = geolib.LE90(z)
    le90_corr = geolib.LE90(z_corr)

    coefs = [ce90, ce90, le90]
    #maxdim = np.ceil(np.max([np.max(np.abs([x, y, z])), ce90, le90]))
    maxdim = np.ceil(np.max([np.percentile(np.abs([x, y, z]), 99), ce90, le90]))
    
    if orthogonal_fig:
        from matplotlib.patches import Ellipse
        #fig_ortho, axa = plt.subplots(1, 3, sharex=True, sharey=True, figsize=(10,5))
        fig_ortho, axa = plt.subplots(1, 3, figsize=(10,5))
        title = 'Co-registration Translation Vector Components, n=%i\n' % x.shape[0]
        title += 'mean: (%0.2f, %0.2f, %0.2f), std: (%0.2f, %0.2f, %0.2f)\n' % (tuple(cmean) + tuple(cstd))
        title += 'med: (%0.2f, %0.2f, %0.2f), nmad: (%0.2f, %0.2f, %0.2f)\n' % (tuple(cmed) + tuple(cnmad))
        title += 'CE90: %0.2f (Bias-corrected: %0.2f), LE90: %0.2f (Bias-corrected: %0.2f)' % (ce90, ce90_corr, le90, le90_corr)
        plt.suptitle(title) 

        dot_prop={'color':'k', 'linestyle':'None', 'marker':'.', 'ms':3, 'label':'ICP correction vector', 'alpha':0.5}
        mean_prop={'color':'r', 'linestyle':'None', 'marker':'o', 'label':'Mean'}

        for ax in axa:
            ax.set_xlim(-maxdim, maxdim)
            ax.set_ylim(-maxdim, maxdim)
            ax.minorticks_on()
            ax.set_aspect('equal')

        axa[0].plot(x, y, **dot_prop)
        axa[0].plot(cmean[0], cmean[1], **mean_prop)
        axa[0].set_xlabel('X offset (m)')
        axa[0].set_ylabel('Y offset (m)')
        e = Ellipse((0,0), 2*ce90, 2*ce90, linewidth=0, alpha=0.1)
        axa[0].add_artist(e)
        axa[0].legend(prop={'size':8}, numpoints=1, loc='upper left')

        axa[1].plot(x, z, **dot_prop)
        axa[1].plot(cmean[0], cmean[2], **mean_prop)
        axa[1].set_xlabel('X offset (m)')
        axa[1].set_ylabel('Z offset (m)')
        e = Ellipse((0,0), 2*ce90, 2*le90, linewidth=0, alpha=0.1)
        axa[1].add_artist(e)

        axa[2].plot(y, z, **dot_prop)
        axa[2].plot(cmean[1], cmean[2], **mean_prop)
        axa[2].set_xlabel('X offset (m)')
        axa[2].set_ylabel('Z offset (m)')
        e = Ellipse((0,0), 2*ce90, 2*le90, linewidth=0, alpha=0.1)
        axa[2].add_artist(e)
        
        plt.tight_layout()

        #Note: postscript doesn't properly handle tansparency
        fig_fn = '%s_translation_vec_local_meters_orthogonal.pdf' % out_fn_prefix
        plt.savefig(fig_fn, dpi=600, bbox_inches='tight')

def make_map(x, y, z, cx, cy):
    f, axa = plt.subplots(3, sharex=True, sharey=True, figsize=(5,10))
    axa[0].set_aspect('equal')
    maxdim = np.ceil(np.percentile(np.abs([x, y, z]), 99))
    #vmin, vmax = (-15, 15)
    vmin, vmax = (-maxdim, maxdim)
    s=5
    cmap='RdYlBu'
    opt={'edgecolor':'k', 'vmin':vmin, 'vmax':vmax, 'cmap':cmap, 's':s, 'lw':0.3}
    sc = axa[0].scatter(cx, cy, c=x, **opt)
    axa[0].set_title("X-offset required to align")
    axa[0].set_aspect('equal')
    axa[1].scatter(cx, cy, c=y, **opt) 
    axa[1].set_title("Y-offset required to align")
    axa[2].scatter(cx, cy, c=z, **opt) 
    axa[2].set_title("Z-offset required to align")
    f.colorbar(sc, ax=axa.ravel().tolist())
    fig_fn = '%s_map.png' % out_fn_prefix
    f.savefig(fig_fn, dpi=300, bbox_inches='tight')

print("Building fn_list")
#fn_list = glob.glob('*dem_align/*align.tif')
fn_list = sys.argv[1:]
fn_list = iolib.fn_list_valid(fn_list)
if not fn_list:
    sys.exit("No valid input files")
print("Isolating x, y, z offsets")
delim='_nuth_'
xyz = np.array([np.array([a[1:] for a in np.array(os.path.split(fn)[-1].split(delim)[-1].split('_'))[0:3]], dtype=float) for fn in fn_list])
print("Extracting center coords")
t_srs = geolib.hma_aea_srs
#t_srs = geolib.conus_aea_srs
#t_srs = geolib.wgs_srs
ll = np.array([geolib.get_center(gdal.Open(fn), t_srs=t_srs) for fn in fn_list])
cy = ll[:,1]
cx = ll[:,0]
m = np.sqrt(np.sum(np.square(xyz), axis=1))

df = pd.DataFrame(xyz, index=fn_list, columns=['x','y','z'])
df['m'] = m
df['cy'] = cy
df['cx'] = cx

df = df.sort_values(by='m', ascending=False)
print(df.shape[0])

if filter:
    print("Correction magnitude")
    stats = malib.print_stats(m)
    print("Outlier magnitude threshold:")
    print(outlier_mag_thresh)
    f=3.5
    print(stats[5]+stats[6]*f)
    idx = (df['m'] > outlier_mag_thresh)
    df[idx].to_csv('%s_bad_fn.txt' % out_fn_prefix, sep=' ', columns='m', index=True, header=False, float_format='%0.2f')

    if mv_bad:
        print("Moving bad solutions")
        import shutil
        baddir = '%s_bad' % out_fn_prefix
        os.makedirs(baddir)
        for i in df[idx].index:
            shutil.move(os.path.split(i)[0], baddir)

    df[~idx].to_csv('%s_good_fn.txt' % out_fn_prefix, sep=' ', columns='m', index=True, header=False)
    df = df[~idx]
    print(df.shape[0])

print("Creating plot")
make_plot3d(df['x'], df['y'], df['z'])
print("Creating map")
make_map(df['x'], df['y'], df['z'], df['cx'], df['cy'])



#! /usr/bin/env python

#David Shean
#10/4/16
#dshean@gmail.com

#Identify DEMs for each site polygon
#Generate DEM stack

import sys
import os
import numpy as np

from osgeo import gdal, ogr, osr

from pygeotools.lib import malib
from pygeotools.lib import geolib
from pygeotools.lib import timelib

if len(sys.argv) != 3:
    sys.exit('Usage is %s sitepoly.shp dempoly.shp' % sys.argv[0])

#This contains polygons defining study areas
#Could also sort glacier polygons by area, then just use largest directly
site_shp_fn = sys.argv[1]
if not os.path.exists(site_shp_fn):
    sys.exit('Unable to find input shp: %s' % site_shp_fn)

site_shp_ds = ogr.Open(site_shp_fn)
site_shp_lyr = site_shp_ds.GetLayer()
site_shp_srs = site_shp_lyr.GetSpatialRef()

#This contains merged shp of outlines for available DEMs
dem_shp_fn = sys.argv[2]
if not os.path.exists(dem_shp_fn):
    sys.exit('Unable to find input shp: %s' % dem_shp_fn)

dem_shp_ds = ogr.Open(dem_shp_fn)
dem_shp_lyr = dem_shp_ds.GetLayer()
dem_shp_srs = dem_shp_lyr.GetSpatialRef()

#outdir = os.path.splitext(shp_fn)[0]+'_stack'
outdir = 'sites_poly'
if not os.path.exists(outdir):
    os.makedirs(outdir)

#Output stack parameters
stack_res = 2 
#stack_res = 8 
buffer = 1000
min_area = 1000000
#UTM 11N
#stack_epsg = 32611
#UTM 10N
stack_epsg = 32610
#This is HMA UTM
#stack_epsg = 32644

dst_srs = osr.SpatialReference()
#dst_srs = geolib.wgs_srs
dst_srs.ImportFromEPSG(stack_epsg)

#site_name_field = 1 
#For 24k glacier polygons
site_name_field = 4 
dem_name_field = 0 

for n,site_feat in enumerate(site_shp_lyr):
    site_name = site_feat.GetFieldAsString(site_name_field) 
    site_name = site_name.replace(" ", "_").lower()
    #Could output json here
    stackdir=os.path.join(outdir,site_name)
    if not os.path.exists(stackdir):
        os.makedirs(stackdir)
    out_fn = os.path.join(stackdir, 'site_%s_%i.csv' % (site_name, stack_epsg))
    f = open(out_fn, 'w')

    site_geom_orig = site_feat.GetGeometryRef()
    site_geom_orig.AssignSpatialReference(site_shp_srs)
    geolib.geom_transform(site_geom_orig, dst_srs)
    
    if buffer is not None:
        site_geom_buff = site_geom_orig.Buffer(buffer)
        site_geom = site_geom_buff
    else:
        site_geom = site_geom_orig

    #This is xmin, xmax, ymin, ymax
    site_extent = site_geom.GetEnvelope()
    stack_extent = [site_extent[0], site_extent[2], site_extent[1], site_extent[3]]
    
    print(site_name)
    print(site_extent)
    print(np.array(geolib.geom_wh(site_geom))/1000.)
    print(site_geom.Area()/1E6)

    #Get output filename from feature, make sure datestr is included
    print(site_name, file=f)
    #print>>f, dst_srs.ExportAsProj4()
    print(site_extent, file=f) 
    
    #Can also set spatial filter for each site
    #https://pcjericks.github.io/py-gdalogr-cookbook/vector_layers.html#spatial-filter

    dem_fn_list = []
    dem_shp_lyr.ResetReading()
    for m,dem_feat in enumerate(dem_shp_lyr):
        dem_name = dem_feat.GetFieldAsString(dem_name_field)
        #Date
        dem_geom = dem_feat.GetGeometryRef()
        dem_geom.AssignSpatialReference(dem_shp_srs)
        geolib.geom_transform(dem_geom, dst_srs)

        #Add interseciton area filter
        igeom = site_geom.Intersection(dem_geom)
        if not igeom.IsEmpty():
            if igeom.Area() > min_area:
                dem_fn_list.append(dem_name)
            #Write out interseciton geom to new file
        #dem_feat.Destroy()

    #Hack to use different DEM resolutions
    #dem_fn_list = ['%s.tif' % i for i in dem_fn_list]
    #dem_fn_list = [i.replace('DEM_32m_trans_32611_100m','DEM_8m_trans.tif') for i in dem_fn_list]
    dem_fn_list = [i.replace('DEM_32m_trans_32611_100m','DEM_%im_trans.tif' % stack_res) for i in dem_fn_list]
    print(dem_fn_list, file=f)
    print("Generating stack")
    s = malib.DEMStack(fn_list=dem_fn_list, outdir=stackdir, res=stack_res, extent=stack_extent, srs=dst_srs, trend=True)

    #temp_lyr = temp_ds.CreateLayer('out', None, ogr.wkbPolygon)
    #temp_lyr.CreateFeature(feat)

    #Mask stack with glacier poly

    dem_dt_list = [timelib.fn_getdatetime(i) for i in dem_fn_list]

    #Isolate summer DEMs
    #dem_fn_list_summer = 
    #Compute seasonal balance each year, composites each spring and each summer, difference
    #See stack_fltr
    #summer = timelib.get_dt_bounds(dem_dt_list, min_rel_dt=(8,15), max_rel_dt=(10,15))
    #spring = timelib.get_dt_bounds(dem_dt_list, min_rel_dt=(4,15), max_rel_dt=(6,15))
    #Alternatively, pick an ideal date, then find everything around
    #summer = timelib.get_closest_dt_padded_idx(datetime(

    #Long-term elevation change NED
    #Clip to glacier extent
    #Stats

    #site_feat.Destroy()
    f = None



#! /usr/bin/env python

import numpy as np
from pygeotools.lib import timelib, geolib, iolib
from osgeo import osr

glas_csv_fn='GLAH14_chad_refdemfilt.csv'
srtm_fn='chad_nasadem_hgt_merge_hgt_adj_proj_hs_az315.tif'
srtm_ds=iolib.fn_getds(srtm_fn)

# dt_ordinal, dt_YYYYMMDD, lat, lon, z_WGS84, z_refdem_med_WGS84, z_refdem_nmad
glas_pts = np.loadtxt(glas_csv_fn, delimiter=',', skiprows=1, dtype=None)

srs=osr.SpatialReference()
srs.ImportFromEPSG(32633)
x, y, z = geolib.cT_helper(glas_pts[:,3], glas_pts[:,2], glas_pts[:,4], geolib.wgs_srs, srs)

#pt_array = glas_pts[:,[3,2,4,0]]
pt_array = np.array([x, y, z, glas_pts[:,0]]).T

#Cluster
dt_thresh=16.0
d = np.diff(pt_array[:,3])
b = np.nonzero(d > dt_thresh)[0] + 1
b = np.hstack((0, b, d.shape[0]))
f_list = []
dt_list = []
for i in range(len(b)-1):
    f = pt_array[b[i]:b[i+1]]
    min_dt = timelib.o2dt(f[:,3].min())
    max_dt = timelib.o2dt(f[:,3].max())
    mid_dt = f[:,3].min() + (f[:,3].max() - f[:,3].min())/2.0
    mean_dt = timelib.o2dt(f[:,3].mean())
    med_dt = timelib.o2dt(np.median(f[:,3]))
    f_list.append(f)
    dt_list.append([min_dt, max_dt, mid_dt, mean_dt, med_dt])

statlist=('median', 'mean', 'count', 'std')
res=300
for n,f in enumerate(f_list):
    mean_dt = dt_list[n][3]
    for stat in statlist:
        g_count, ds = geolib.block_stats_grid_gen(f[:,0], f[:,1], f[:,2], res=res, srs=srs, stat=stat)
        out_fn = mean_dt.strftime('%Y%m%d_%H%M_')+stat+'_'+str(res)+'m.tif'
        iolib.writeGTiff(g_count, out_fn, ds)

#Make stacks



#! /usr/bin/env python

#Todo
#Better outlier removal
#Check Nuth and Kaab bin median
#Implement check for empty diff

import sys
import os
import argparse
import subprocess

from osgeo import gdal
import numpy as np
import matplotlib.pyplot as plt

from pygeotools.lib import iolib, malib, geolib, warplib, filtlib

from demcoreg import coreglib, dem_mask

from imview.lib import pltlib

def get_mask(ds, mask_list, dem_fn=None):
    #This returns True (1) for areas to mask, False (0) for valid static surfaces
    static_mask = dem_mask.get_mask(ds, mask_list, dem_fn, writeout=False)
    #return ~(static_mask)
    return static_mask

def compute_offset(ref_dem_ds, src_dem_ds, src_dem_fn, mode='nuth', remove_outliers=True, max_offset=100, \
        max_dz=50, slope_lim=(0.1, 40), mask_list=['glaciers',], plot=True):
    #Make sure the input datasets have the same resolution/extent
    #Use projection of source DEM
    ref_dem_clip_ds, src_dem_clip_ds = warplib.memwarp_multi([ref_dem_ds, src_dem_ds], \
            res='max', extent='intersection', t_srs=src_dem_ds, r='cubic')

    #Compute size of NCC and SAD search window in pixels
    res = float(geolib.get_res(ref_dem_clip_ds, square=True)[0])
    max_offset_px = (max_offset/res) + 1
    #print(max_offset_px)
    pad = (int(max_offset_px), int(max_offset_px))

    #This will be updated geotransform for src_dem
    src_dem_gt = np.array(src_dem_clip_ds.GetGeoTransform())

    #Load the arrays
    ref_dem = iolib.ds_getma(ref_dem_clip_ds, 1)
    src_dem = iolib.ds_getma(src_dem_clip_ds, 1)

    print("Elevation difference stats for uncorrected input DEMs (src - ref)")
    diff_euler = src_dem - ref_dem

    static_mask = get_mask(src_dem_clip_ds, mask_list, src_dem_fn)
    diff_euler = np.ma.array(diff_euler, mask=static_mask)

    if diff_euler.count() == 0:
        sys.exit("No overlapping, unmasked pixels shared between input DEMs")

    #This needs further testing
    if remove_outliers:
        print("Removing outliers")
        print("Initial pixel count:")
        print(diff_euler.count())

        print("Absolute dz filter: %0.2f" % max_dz)
        #Absolute dz filter
        diff_euler = np.ma.masked_greater(diff_euler, max_dz)
        print(diff_euler.count())

        #Outlier dz filter
        f = 3
        nmad, med = malib.mad(diff_euler, return_med=True)
        rmin = med - f*nmad
        rmax = med + f*nmad
        print("3-sigma filter: %0.2f - %0.2f" % (rmin, rmax))
        diff_euler = np.ma.masked_outside(diff_euler, rmin, rmax)
        print(diff_euler.count())

        #Should also apply to original ref_dem and src_dem for sad and ncc

    #Generate slope map
    #Want to use higher quality DEM, should determine automatically from original res/count
    #slope = geolib.gdaldem_mem_ds(ref_dem_clip_ds, processing='slope', returnma=True, computeEdges=False)
    slope = geolib.gdaldem_mem_ds(src_dem_clip_ds, processing='slope', returnma=True, computeEdges=False)

    #slope_stats = malib.print_stats(slope)
    #Remove extreme slopes
    print("Slope filter: %0.2f - %0.2f" % slope_lim)
    print("Initial count: %i" % slope.count()) 
    slope = filtlib.range_fltr(slope, slope_lim) 
    print(slope.count())

    print("Computing aspect")
    #aspect = geolib.gdaldem_mem_ds(ref_dem_clip_ds, processing='aspect', returnma=True, computeEdges=False)
    aspect = geolib.gdaldem_mem_ds(src_dem_clip_ds, processing='aspect', returnma=True, computeEdges=False)

    ref_dem_clip_ds = None
    src_dem_clip_ds = None

    #Apply slope filter to diff_euler
    #Note that we combine masks from diff_euler and slope in coreglib
    diff_euler = np.ma.array(diff_euler, mask=np.ma.getmaskarray(slope))

    #Get final mask after filtering
    static_mask = np.ma.getmaskarray(diff_euler)

    #Compute stats for new masked difference map
    print("Filtered difference map")
    diff_stats = malib.print_stats(diff_euler)
    dz = diff_stats[5]

    print("Computing sub-pixel offset between DEMs using mode: %s" % mode)

    #By default, don't create output figure
    fig = None

    #Default horizntal shift is (0,0)
    dx = 0
    dy = 0

    #Sum of absolute differences
    if mode == "sad":
        ref_dem = np.ma.array(ref_dem, mask=static_mask)
        src_dem = np.ma.array(src_dem, mask=static_mask)
        m, int_offset, sp_offset = coreglib.compute_offset_sad(ref_dem, src_dem, pad=pad)
        #Geotransform has negative y resolution, so don't need negative sign
        #np array is positive down
        #GDAL coordinates are positive up
        dx = sp_offset[1]*src_dem_gt[1]
        dy = sp_offset[0]*src_dem_gt[5]
    #Normalized cross-correlation of clipped, overlapping areas
    elif mode == "ncc":
        ref_dem = np.ma.array(ref_dem, mask=static_mask)
        src_dem = np.ma.array(src_dem, mask=static_mask)
        m, int_offset, sp_offset, fig = coreglib.compute_offset_ncc(ref_dem, src_dem, \
                pad=pad, prefilter=False, plot=plot)
        dx = sp_offset[1]*src_dem_gt[1]
        dy = sp_offset[0]*src_dem_gt[5]
    #Nuth and Kaab (2011)
    elif mode == "nuth":
        #Compute relationship between elevation difference, slope and aspect
        fit_param, fig = coreglib.compute_offset_nuth(diff_euler, slope, aspect, plot=plot)
        if fit_param is None:
            print("Failed to calculate horizontal shift")
        else:
            #fit_param[0] is magnitude of shift vector
            #fit_param[1] is direction of shift vector
            #fit_param[2] is mean bias divided by tangent of mean slope
            #print(fit_param)
            dx = fit_param[0]*np.sin(np.deg2rad(fit_param[1]))
            dy = fit_param[0]*np.cos(np.deg2rad(fit_param[1]))
            med_slope = malib.fast_median(slope)
            nuth_dz = fit_param[2]*np.tan(np.deg2rad(med_slope))
            print('Median dz: %0.2f\nNuth dz: %0.2f' % (dz, nuth_dz))
            #dz = nuth_dz
    elif mode == "all":
        print("Not yet implemented")
        #Want to compare all methods, average offsets
        #m, int_offset, sp_offset = coreglib.compute_offset_sad(ref_dem, src_dem)
        #m, int_offset, sp_offset = coreglib.compute_offset_ncc(ref_dem, src_dem)
    elif mode == "none":
        print("Skipping alignment, writing out DEM with median bias over static surfaces removed")
        dst_fn = outprefix+'_med%0.1f.tif' % dz
        iolib.writeGTiff(src_dem_orig + dz, dst_fn, src_dem_ds)
        sys.exit()
    #Note: minus signs here since we are computing dz=(src-ref), but adjusting src
    return -dx, -dy, -dz, static_mask, fig

def getparser():
    parser = argparse.ArgumentParser(description="Perform DEM co-registration using multiple algorithms", \
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('ref_fn', type=str, help='Reference DEM filename')
    parser.add_argument('src_fn', type=str, help='Source DEM filename to be shifted')
    parser.add_argument('-mode', type=str, default='nuth', choices=['ncc', 'sad', 'nuth', 'none'], \
            help='Type of co-registration to use')
    parser.add_argument('-mask_list', nargs='+', type=str, default=['glaciers',], choices=dem_mask.mask_choices, \
            help='Define masks to use to limit reference surfaces for co-registration')
    parser.add_argument('-tiltcorr', action='store_true', \
            help='After preliminary translation, fit plane to residual elevation offsets and remove')
    parser.add_argument('-tol', type=float, default=0.02, \
            help='When iterative translation magnitude is below this tolerance (meters), break and write out corrected DEM')
    parser.add_argument('-max_offset', type=float, default=100, \
            help='Maximum expected horizontal offset in meters')
    parser.add_argument('-max_dz', type=float, default=100, \
            help='Maximum expected vertical offset in meters, used to filter outliers')
    parser.add_argument('-slope_lim', type=float, nargs=2, default=(0.1, 40), \
            help='Minimum and maximum surface slope limits to consider')
    parser.add_argument('-max_iter', type=int, default=20, \
            help='Maximum number of iterations, if tol is not reached')
    parser.add_argument('-outdir', default=None, help='Output directory')
    return parser

#Defined a second main to allow recursion with new arguments for second run
#Likely a cleaner strategy for this
def main2(args):
    #Should check that files exist
    ref_dem_fn = args.ref_fn
    src_dem_fn = args.src_fn

    mode = args.mode
    mask_list = args.mask_list
    max_offset = args.max_offset
    max_dz = args.max_dz
    slope_lim = args.slope_lim
    tiltcorr = args.tiltcorr

    #These are tolerances (in meters) to stop iteration
    tol = args.tol
    min_dx = tol
    min_dy = tol
    min_dz = tol

    #Maximum number of iterations
    max_n = args.max_iter

    outdir = args.outdir
    if outdir is None:
        #outdir = os.path.splitext(src_dem_fn)[0] + '_dem_align_lt1.5m_err'
        outdir = os.path.splitext(src_dem_fn)[0] + '_dem_align'

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    outprefix = '%s_%s' % (os.path.splitext(os.path.split(src_dem_fn)[-1])[0], \
            os.path.splitext(os.path.split(ref_dem_fn)[-1])[0])
    outprefix = os.path.join(outdir, outprefix)

    print("\nReference: %s" % ref_dem_fn)
    print("Source: %s" % src_dem_fn)
    print("Mode: %s" % mode)
    print("Output: %s\n" % outprefix)

    src_dem_ds = gdal.Open(src_dem_fn)
    #Create a copy to be updated in place
    #src_dem_ds_align = iolib.mem_drv.CreateCopy('', src_dem_ds, 0)
    #This is another approach that duplicates CreateCopy functionality
    ##src_dem_ds_align = geolib.mem_ds_copy(src_dem_ds)

    #Get local cartesian coordinate system
    #local_srs = geolib.localortho_ds(src_dem_ds)
    local_srs = geolib.localtmerc_ds(src_dem_ds)
    #local_srs = geolib.ds_get_srs(src_dem_ds)

    src_dem_ds_align = warplib.memwarp_multi_fn([src_dem_fn,], t_srs=local_srs, r='cubic')[0]

    src_dem_ds = None
    
    #Resample to match "source" DEM
    ref_dem_ds = warplib.memwarp_multi_fn([ref_dem_fn,], res=src_dem_ds_align, extent=src_dem_ds_align, t_srs=local_srs, r='cubic')[0]
    #ref_dem_ds = gdal.Open(ref_dem_fn) 

    #Iteration number
    n = 1
    #Cumulative offsets
    dx_total = 0
    dy_total = 0
    dz_total = 0

    #Now iteratively update geotransform and vertical shift
    while True:
        print("*** Iteration %i ***" % n)
        dx, dy, dz, static_mask, fig = compute_offset(ref_dem_ds, src_dem_ds_align, src_dem_fn, mode, max_offset, \
                mask_list=mask_list, max_dz=max_dz, slope_lim=slope_lim, plot=True)
        xyz_shift_str_iter = "dx=%+0.2fm, dy=%+0.2fm, dz=%+0.2fm" % (dx, dy, dz)
        print("Incremental offset: %s" % xyz_shift_str_iter)

        #Should make an animation of this converging
        if n == 1: 
            static_mask_orig = static_mask
            if fig is not None:
                dst_fn = outprefix + '_%s_iter%02i_plot.png' % (mode, n)
                print("Writing offset plot: %s" % dst_fn)
                fig.gca().set_title(xyz_shift_str_iter)
                fig.savefig(dst_fn, dpi=300)

        #Apply the horizontal shift to the original dataset
        src_dem_ds_align = coreglib.apply_xy_shift(src_dem_ds_align, dx, dy, createcopy=False)
        #Should 
        src_dem_ds_align = coreglib.apply_z_shift(src_dem_ds_align, dz, createcopy=False)

        dx_total += dx
        dy_total += dy
        dz_total += dz
        print("Cumulative offset: dx=%+0.2fm, dy=%+0.2fm, dz=%+0.2fm" % (dx_total, dy_total, dz_total))

        n += 1
        print("\n")
        #If magnitude of shift in all directions is less than tol
        #if n > max_n or (abs(dx) <= min_dx and abs(dy) <= min_dy and abs(dz) <= min_dz):
        #If magnitude of shift is less than tol
        dm = np.sqrt(dx**2 + dy**2 + dz**2)
        if n > max_n or dm < tol:
            if fig is not None:
                dst_fn = outprefix + '_%s_iter%02i_plot.png' % (mode, n)
                print("Writing offset plot: %s" % dst_fn)
                fig.gca().set_title(xyz_shift_str_iter)
                fig.savefig(dst_fn, dpi=300)
            break

    #String to append to output filenames
    xyz_shift_str_cum = '_%s_x%+0.2f_y%+0.2f_z%+0.2f' % (mode, dx_total, dy_total, dz_total)
    if tiltcorr:
        xyz_shift_str_cum += "_tiltcorr"

    #Compute original elevation difference
    if True:
        src_dem_ds = gdal.Open(src_dem_fn)
        ref_dem_clip_ds, src_dem_clip_ds = warplib.memwarp_multi([ref_dem_ds, src_dem_ds], \
                res='max', extent='intersection', t_srs=local_srs, r='cubic')
        src_dem_ds = None
        ref_dem_orig = iolib.ds_getma(ref_dem_clip_ds)
        src_dem_orig = iolib.ds_getma(src_dem_clip_ds)
        #Needed for plotting
        ref_dem_hs = geolib.gdaldem_mem_ds(ref_dem_clip_ds, processing='hillshade', returnma=True, computeEdges=True)
        src_dem_hs = geolib.gdaldem_mem_ds(src_dem_clip_ds, processing='hillshade', returnma=True, computeEdges=True)
        res = float(geolib.get_res(ref_dem_clip_ds, square=True)[0])
        diff_euler_orig = src_dem_orig - ref_dem_orig
        #Only compute stats over valid surfaces
        static_mask_orig = get_mask(src_dem_clip_ds, mask_list, src_dem_fn)
        #Note: this doesn't include outlier removal or slope mask!
        static_mask_orig = np.logical_or(np.ma.getmaskarray(diff_euler_orig), static_mask_orig)
        #For some reason, ASTER DEM diff have a spike near the 0 bin, could be an issue with masking?
        diff_euler_orig_compressed = diff_euler_orig[~static_mask_orig].compressed()
        diff_euler_orig_stats = np.array(malib.print_stats(diff_euler_orig_compressed))

        #Write out original eulerian difference map
        print("Writing out original euler difference map for common intersection before alignment")
        orig_eul_fn = outprefix + '_orig_dz_eul.tif'
        iolib.writeGTiff(diff_euler_orig, orig_eul_fn, ref_dem_clip_ds)
        src_dem_clip_ds = None
        ref_dem_clip_ds = None

    #Compute final elevation difference
    if True:
        ref_dem_clip_ds_align, src_dem_clip_ds_align = warplib.memwarp_multi([ref_dem_ds, src_dem_ds_align], \
                res='max', extent='intersection', t_srs=local_srs, r='cubic')
        ref_dem_align = iolib.ds_getma(ref_dem_clip_ds_align, 1)
        src_dem_align = iolib.ds_getma(src_dem_clip_ds_align, 1)
        #Need this for scalebar plot
        #res = float(geolib.get_res(src_dem_clip_ds_align, square=True)[0])
        diff_euler_align = src_dem_align - ref_dem_align
        src_dem_align = None
        ref_dem_align = None
        #Get updated, final mask
        static_mask_final = get_mask(src_dem_clip_ds_align, mask_list, src_dem_fn)
        src_dem_clip_ds_align = None
        static_mask_final = np.logical_or(np.ma.getmaskarray(diff_euler_align), static_mask_final)
        diff_euler_align_compressed = diff_euler_align[~static_mask_final].compressed()
        diff_euler_align_stats = np.array(malib.print_stats(diff_euler_align_compressed))

        #Fit plane to residuals and remove
        #Compute higher-order fits?
        #Could also attempt to model along-track and cross-track artifacts
        if tiltcorr:
            print("Applying planar tilt correction")
            gt = ref_dem_clip_ds_align.GetGeoTransform()
            #Need to apply the mask here, so we're only fitting over static surfaces
            #Note that the origmask=False will compute vals for all x and y indices, which is what we want
            #Should offer option for polynomial of arbitrary order
            #Also, want a better robust fit - maybe throw out more outliers
            vals, resid, coeff = geolib.ma_fitplane(np.ma.array(diff_euler_align, mask=static_mask_final), \
                    gt, perc=(12.5, 87.5), origmask=False)
            #Remove planar offset from difference map
            diff_euler_align -= vals
            #Remove planar offset from aligned src_dem
            #Note: dimensions of ds and vals will be different as vals are computed for clipped intersection
            #Recompute planar offset for src_dem_ds_align extent
            xgrid, ygrid = geolib.get_xy_grids(src_dem_ds_align)
            vals = coeff[0]*xgrid + coeff[1]*ygrid + coeff[2]
            src_dem_ds_align = coreglib.apply_z_shift(src_dem_ds_align, -vals, createcopy=False)
            diff_euler_align_compressed = diff_euler_align[~static_mask_final].compressed()
            diff_euler_align_stats = np.array(malib.print_stats(diff_euler_align_compressed))
            if True:
                print("Creating plot of planar correction")
                fig, ax = plt.subplots(figsize=(6, 6))
                fitplane_clim = malib.calcperc(vals, (2,98))
                im = ax.imshow(vals, cmap='cpt_rainbow', clim=fitplane_clim)
                pltlib.add_scalebar(ax, res=res)
                pltlib.hide_ticks(ax)
                pltlib.add_cbar(ax, im, arr=vals, clim=fitplane_clim, label='Fit plane residuals (m)')
                fig.tight_layout()
                tiltcorr_fig_fn = outprefix + '%s_align_dz_eul_fitplane.png' % xyz_shift_str_cum
                print("Writing out figure: %s" % tiltcorr_fig_fn)
                fig.savefig(tiltcorr_fig_fn, dpi=300)

        #Write out aligned eulerian difference map for clipped extent with vertial offset removed
        align_eul_fn = outprefix + '%s_align_dz_eul.tif' % xyz_shift_str_cum
        print("Writing out aligned difference map with median vertical offset removed")
        iolib.writeGTiff(diff_euler_align, align_eul_fn, ref_dem_clip_ds_align)
        ref_dem_clip_ds_align = None

    #Write out aligned dem_2 with vertial offset removed
    if True:
        align_fn = outprefix + '%s_align.tif' % xyz_shift_str_cum
        print("Writing out shifted src_dem with median vertical offset removed: %s" % align_fn)
        #Might be cleaner way to write out MEM ds directly to disk
        src_dem_align = iolib.ds_getma(src_dem_ds_align)
        iolib.writeGTiff(src_dem_align, align_fn, src_dem_ds_align)
        src_dem_align = None
        src_dem_ds_align = None

    #Create output plot
    if True:
        print("Creating final plot")
        f, axa = plt.subplots(2, 3, figsize=(11, 8.5))
        for ax in axa.ravel()[:-1]:
            ax.set_facecolor('k')
            pltlib.hide_ticks(ax)
        dem_clim = malib.calcperc(ref_dem_orig, (2,98))
        axa[0,0].imshow(ref_dem_hs, cmap='gray')
        im = axa[0,0].imshow(ref_dem_orig, cmap='cpt_rainbow', clim=dem_clim, alpha=0.6)
        pltlib.add_cbar(axa[0,0], im, arr=ref_dem_orig, clim=dem_clim, label=None)
        pltlib.add_scalebar(axa[0,0], res=res)
        axa[0,0].set_title('Reference DEM')
        axa[0,1].imshow(src_dem_hs, cmap='gray')
        im = axa[0,1].imshow(src_dem_orig, cmap='cpt_rainbow', clim=dem_clim, alpha=0.6)
        pltlib.add_cbar(axa[0,1], im, arr=src_dem_orig, clim=dem_clim, label=None)
        axa[0,1].set_title('Source DEM')
        #axa[0,2].imshow(~static_mask_orig, clim=(0,1), cmap='gray')
        axa[0,2].imshow(~static_mask, clim=(0,1), cmap='gray')
        axa[0,2].set_title('Surfaces for co-registration')
        dz_clim = malib.calcperc_sym(diff_euler_orig_compressed, (5, 95))
        im = axa[1,0].imshow(diff_euler_orig, cmap='RdBu', clim=dz_clim)
        pltlib.add_cbar(axa[1,0], im, arr=diff_euler_orig, clim=dz_clim, label=None)
        axa[1,0].set_title('Elev. Diff. Before (m)')
        im = axa[1,1].imshow(diff_euler_align, cmap='RdBu', clim=dz_clim)
        pltlib.add_cbar(axa[1,1], im, arr=diff_euler_align, clim=dz_clim, label=None)
        axa[1,1].set_title('Elev. Diff. After (m)')

        #Tried to insert Nuth fig here
        #ax_nuth.change_geometry(1,2,1)
        #f.axes.append(ax_nuth)

        bins = np.linspace(dz_clim[0], dz_clim[1], 128)
        axa[1,2].hist(diff_euler_orig_compressed, bins, color='g', label='Before', alpha=0.5)
        axa[1,2].hist(diff_euler_align_compressed, bins, color='b', label='After', alpha=0.5)
        axa[1,2].axvline(0, color='k', linewidth=0.5, linestyle=':')
        axa[1,2].set_xlabel('Elev. Diff. (m)')
        axa[1,2].set_ylabel('Count (px)')
        axa[1,2].set_title("Source - Reference")
        #axa[1,2].legend(loc='upper right')
        #before_str = 'Before\nmean: %0.2f\nstd: %0.2f\nmed: %0.2f\nnmad: %0.2f' % tuple(diff_euler_orig_stats[np.array((3,4,5,6))])
        #after_str = 'After\nmean: %0.2f\nstd: %0.2f\nmed: %0.2f\nnmad: %0.2f' % tuple(diff_euler_align_stats[np.array((3,4,5,6))])
        before_str = 'Before\nmed: %0.2f\nnmad: %0.2f' % tuple(diff_euler_orig_stats[np.array((5,6))])
        axa[1,2].text(0.05, 0.95, before_str, va='top', color='g', transform=axa[1,2].transAxes)
        after_str = 'After\nmed: %0.2f\nnmad: %0.2f' % tuple(diff_euler_align_stats[np.array((5,6))])
        axa[1,2].text(0.65, 0.95, after_str, va='top', color='b', transform=axa[1,2].transAxes)

        suptitle = '%s\nx: %+0.2fm, y: %+0.2fm, z: %+0.2fm' % (os.path.split(outprefix)[-1], dx_total, dy_total, dz_total)
        f.suptitle(suptitle)
        f.tight_layout()
        plt.subplots_adjust(top=0.90)

        fig_fn = outprefix + '%s_align.png' % xyz_shift_str_cum
        print("Writing out figure: %s" % fig_fn)
        f.savefig(fig_fn, dpi=300)

    #Removing residual planar tilt can introduce additional slope/aspect dependent offset
    #Want to run another round of main dem_align after removing planar tilt
    if tiltcorr:
        print("\n Rerunning after applying tilt correction \n")
        #Create copy of original arguments
        import copy
        args2 = copy.copy(args)
        #Use aligned, tilt-corrected DEM as input src_fn for second round
        args2.src_fn = align_fn 
        #Assume we've already corrected most of the tilt during first round (also prevents endless loop)
        args2.tiltcorr = False
        main2(args2)

def main(argv=None):
    parser = getparser()
    args = parser.parse_args()
    main2(args)

if __name__ == "__main__":
    main()



#! /usr/bin/env python

"""
Utility to sample input raster for a set of input xyz points and print statistics
"""

import sys
import os
import csv    

from osgeo import gdal
import numpy as np
from pygeotools.lib import geolib, iolib, malib

r_fn = sys.argv[1]
pt_fn = sys.argv[2]

#In the future, allow user to specify point_srs, fields
#For now, assume output is from filter_glas.py, coords are lat/lon
xy_srs = geolib.wgs_srs
#Field numbers of x y z coordinates
xcol = 3
ycol = 2
zcol = 4

print("Loading points: %s" % pt_fn)
#This needs to return a header
pts = iolib.readcsv(pt_fn)

#Assume that all of our points are preselected to fall withing raster extent
#Can add option for spatial filter, see filter_glas.py

r_ds = gdal.Open(r_fn)

print("\nInput raster: %s" % r_fn)
print("Input points: %s" % pt_fn)
print("\nSampling %i points\n" % pts.shape[0])

#This returns median and mad arrays for all values within pad
#Should add option to return number of pixels in sample 
#Use 'glas' here to sample 70 m spot
samp = geolib.sample(r_ds, pts[:,xcol], pts[:,ycol], xy_srs=xy_srs, pad='glas', count=True)
samp_idx = ~(np.ma.getmaskarray(samp[:,0]))
nsamp = samp_idx.nonzero()[0].size

if nsamp == 0:
    sys.exit("No valid samples")
else:
    pts_mask = np.ma.hstack([pts[samp_idx], samp[samp_idx], (pts[samp_idx,zcol] - samp[samp_idx,0])[:,np.newaxis]])

    print("Sample difference (raster - point) statistics:")
    malib.print_stats(pts_mask[:,-1])

    """
    #Update header
    out_hdr_str = None
    if hdr is not None:
        out_hdr = hdr + ['samp_med','samp_nmad','samp_count']
        out_hdr_str = ', '.join(fieldnames)

    #Format for new cols
    #fmt = ['%0.2f', '%0.2f', '%i']
    #fmt = None
    """

    out_pt_fn = os.path.splitext(r_fn)[0]+'_'+os.path.splitext(os.path.split(pt_fn)[-1])[0]+'_sample.csv'
    print("\nWriting out %i points with valid samples:\n%s\n" % (nsamp, out_pt_fn))
    #np.savetxt(out_pt_fn, pts_mask, fmt=fmt, delimiter=',', header=out_hdr_str)
    np.savetxt(out_pt_fn, pts_mask, delimiter=',')

#Should migrate robust_stats functionality to malib
#robust_stats.py -col -1 out_pt_fn



#! /usr/bin/env python

#David Shean
#dshean@gmail.com

#Perform error analysis for DEM output from pc_align

import sys
import os
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from pygeotools.lib import timelib, geolib, malib

#shell
#filenames=$(ls WV*/dem*/*align/*-DEM.tif)
#error_analysis.py $filenames

#ipython
#filenames = !ls *align/*-DEM.tif
#run ~/src/demtools/error_analysis.py $filenames.s

#Compute standard error for all DEMs
#SE=SD/sqrt(n)

def shift_ll2proj(fn, llz):
    from osgeo import gdal, osr
    from pygeotools.lib import geolib
    ds = gdal.Open(fn)
    s_srs = geolib.wgs_srs
    t_srs = geolib.get_ds_srs(ds)
    shift = None
    if t_srs is not None and not s_srs.IsSame(t_srs):
        #center is lon, lat
        #llz is lat, lon
        c = geolib.get_center(ds, t_srs=s_srs)
        c_shift = [c[0]+llz[1], c[1]+llz[0]]
        ct = osr.CoordinateTransformation(s_srs, t_srs)
        c_proj = list(ct.TransformPoint(*c)[0:2])
        c_shift_proj = list(ct.TransformPoint(*c_shift)[0:2])
        shift = list([c_shift_proj[0] - c_proj[0], c_shift_proj[1] - c_proj[1]])
        shift.append(llz[2])
    return shift

def parse_pc_align_log(fn):
    import re
    error_dict = None
    #Determine log filename
    import glob
    log_fn = glob.glob(fn.rsplit('-DEM', 1)[0]+'*.log')
    if not log_fn:
        log_fn = glob.glob(fn.rsplit('-DEM', 1)[0]+'*align/*.log')

    if not log_fn:
        print "Failed to locate align log for %s" % fn
    else:
        log_fn = log_fn[0]
        print(log_fn)
        f = open(log_fn)

        error_dict = {}
        error_dict['File'] = fn
        error_dict['Date'] = timelib.fn_getdatetime(fn)

        #This handles cases where no sampling was performed
        error_dict['Input Sampled 16th Percentile Error'] = np.nan 
        error_dict['Input Sampled Median Error'] = np.nan 
        error_dict['Input Sampled 84th Percentile Error'] = np.nan 
        error_dict['Input Sampled Error Spread'] = np.nan 
        error_dict['Output Sampled 16th Percentile Error'] = np.nan 
        error_dict['Output Sampled Median Error'] = np.nan 
        error_dict['Output Sampled 84th Percentile Error'] = np.nan 
        error_dict['Output Sampled Error Spread'] = np.nan 
        #error_dict['Translation vector (North-East-Down, meters)'] = [np.nan, np.nan, np.nan]

        #Set default reference type to point
        error_dict['Ref type'] = 'point'

        temp = []
        for line in f:
            key = 'Loaded points'
            if key in line:
                temp.append(int(re.split(':', line.rstrip())[1]))
            key = 'Number of errors'
            if key in line:
                error_dict[key] = int(re.split(':', line.rstrip())[1])
            key = 'Input: error percentile'
            if key in line:
                line_a = re.split(': |, ', line.rstrip())
                error_dict['Input 16th Percentile Error'] = float(line_a[3])
                error_dict['Input Median Error'] = float(line_a[5])
                error_dict['Input 84th Percentile Error'] = float(line_a[7])
            """
            key = 'Input: error mean'
            if key in line:
                line_a = re.split(': |, ', line.rstrip())
                error_dict['Input Mean Error'] = float(line_a[2])
                error_dict['Input Std Error'] = float(line_a[4])
            """
            #This pulls the line 
            #Input: mean of smallest errors: 25%: 7.82061, 50%: 9.71931, 75%: 10.9917, 100%: 12.2715
            #Want the final value
            key = 'Input: mean'
            if key in line:
                line_a = re.split(': |, ', line.rstrip())
                error_dict['Input Mean Error'] = float(line_a[-1])
            key = 'Output: error percentile'
            if key in line:
                line_a = re.split(': |, ', line.rstrip())
                error_dict['Output 16th Percentile Error'] = float(line_a[3])
                error_dict['Output Median Error'] = float(line_a[5])
                error_dict['Output 84th Percentile Error'] = float(line_a[7])
            """
            key = 'Output: error mean'
            if key in line:
                line_a = re.split(': |, ', line.rstrip())
                error_dict['Output Mean Error'] = float(line_a[2])
                error_dict['Output Std Error'] = float(line_a[4])
            """
            key = 'Output: mean'
            if key in line:
                line_a = re.split(': |, ', line.rstrip())
                error_dict['Output Mean Error'] = float(line_a[-1])
            key = 'Translation vector (Cartesian, meters)'
            #Previous versions of pc_align output this
            #key = 'Translation vector (meters)'
            if key in line:
                error_dict['Translation vector (Cartesian, meters)'] = list(float(i) for i in re.split('Vector3\(', line.rstrip())[1][:-1].split(',')) 
                #error_dict['Translation vector (meters)'] = list(float(i) for i in re.split('Vector3\(', line.rstrip())[1][:-1].split(',')) 
            key = 'Translation vector (North-East-Down, meters)'
            if key in line:
                error_dict['Translation vector (North-East-Down, meters)'] = list(float(i) for i in re.split('Vector3\(', line.rstrip())[1][:-1].split(',')) 
            key = 'Translation vector magnitude (meters)'
            if key in line:
                error_dict[key] = float(re.split(':', line.rstrip())[1])
            key = 'Translation vector (lat,lon,z)'
            if key in line:
                error_dict[key] = list(float(i) for i in re.split('Vector3\(', line.rstrip())[1][:-1].split(',')) 
                shift_proj = shift_ll2proj(fn, error_dict[key])
                key = 'Translation vector (Proj meters)'
                error_dict[key] = shift_proj

            #This is the output from the point sampling post-alignment
            key = 'Error percentiles'
            if key in line:
                #This is a hack to capture both sampling of input and output
                if 'Output Sampled 16th Percentile Error' in error_dict:
                    error_dict['Input Sampled 16th Percentile Error'] = error_dict['Output Sampled 16th Percentile Error']
                    error_dict['Input Sampled Median Error'] = error_dict['Output Sampled Median Error']
                    error_dict['Input Sampled 84th Percentile Error'] = error_dict['Output Sampled 84th Percentile Error']
                    error_dict['Input Sampled Error Spread'] = error_dict['Output Sampled Error Spread']
                line_a = re.split(': |, ', line.rstrip())
                error_dict['Output Sampled 16th Percentile Error'] = float(line_a[2])
                error_dict['Output Sampled Median Error'] = float(line_a[4])
                error_dict['Output Sampled 84th Percentile Error'] = float(line_a[6])
                error_dict['Output Sampled Error Spread'] = float(line_a[6]) - float(line_a[2])
            #key = 'compute_dh'
            #Note: these are not computed for absolute values by compute_dh
            key = 'count:'
            if key in line:
                error_dict['Ref type'] = 'grid'
                #This is a hack to capture both sampling of input and output
                if 'Output Sampled 16th Percentile Error' in error_dict:
                    error_dict['Input Sampled 16th Percentile Error'] = error_dict['Output Sampled 16th Percentile Error']
                    error_dict['Input Sampled Median Error'] = error_dict['Output Sampled Median Error']
                    error_dict['Input Sampled 84th Percentile Error'] = error_dict['Output Sampled 84th Percentile Error']
                    error_dict['Input Sampled Error Spread'] = error_dict['Output Sampled Error Spread']
                #Assume the following format for stats:
                #count: 349835 min: -51.39 max: 22.00 mean: 0.29 std: 0.49 med: 0.28 mad: 0.37 \
                #q1: 0.04 q2: 0.54 iqr: 0.50 mode: 0.29 p16: -0.07 p84: 0.66 spread: 0.37
                line_a = re.split(': | ', line.rstrip())
                error_dict['Output Sampled 16th Percentile Error'] = float(line_a[23])
                error_dict['Output Sampled Median Error'] = float(line_a[11])
                error_dict['Output Sampled 84th Percentile Error'] = float(line_a[25])
                error_dict['Output Sampled Error Spread'] = float(line_a[25]) - float(line_a[23])
            key = 'Mean error'
            if key in line:
                if 'Output Sampled Mean Error' in error_dict:
                    error_dict['Input Sampled Mean Error'] = error_dict['Output Sampled Mean Error']
                error_dict['Output Sampled Mean Error'] = float(re.split(':', line.rstrip())[1])
            key = 'RMSE'
            if key in line:
                if 'Output Sampled RMSE' in error_dict:
                    error_dict['Input Sampled RMSE'] = error_dict['Output Sampled RMSE']
                error_dict['Output Sampled RMSE'] = float(re.split(':', line.rstrip())[1])
            key = 'Absolute Median Error'
            if key in line:
                if 'Output Absolute Median Error' in error_dict:
                    error_dict['Input Absolute Median Error'] = error_dict['Output Absolute Median Error']
                error_dict['Output Absolute Median Error'] = float(re.split(':', line.rstrip())[1])
                
        error_dict['Source points'] = temp[0] 
        error_dict['Reference points'] = temp[1] 
    
    return error_dict

def make_plot(x,y,yerr=None,c='k',ms=4,label=None,abs=False):
    y_mean = y.mean()
    y_std = y.std()
    y_med = np.ma.median(y)
    y_nmad = malib.mad(y)
    #plt.plot(x, y, label=label, color=c, marker='o', linestyle='None')
    plt.scatter(x, y, label=label, color=c, marker='o', s=ms)
    if yerr is not None:
        plt.errorbar(x, y, yerr=yerr, color=c, linestyle='None', elinewidth=0.5, capsize=np.sqrt(ms), alpha=0.5)
    plt.axhline(y_med, color=c, linestyle='--', alpha=0.5)
    plt.axhline(y_med + y_nmad, color=c, linestyle=':', alpha=0.5)
    plt.axhline(y_med - y_nmad, color=c, linestyle=':', alpha=0.5)
    plt.axhline(0, color='k', linewidth=0.5, linestyle='-', alpha=0.5)
    ax = plt.gca()
    plt.minorticks_on()
    #ax.tick_params(axis='y',which='minor',left='on')
    if abs:
        ax.set_ylim(bottom=0.0)

#Draw ellipsoid in 3D plot
#http://stackoverflow.com/questions/7819498/plotting-ellipsoid-with-matplotlib
def make_plot3d(x, y, z, title=None, orthogonal_fig=True):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.set_aspect('equal')
    ax.set_xlabel('X offset (m)')
    ax.set_ylabel('Y offset (m)')
    ax.set_zlabel('Z offset (m)')
    if title is not None:
        plt.suptitle(title)
    ax.plot(x, y, z, 'o')

    cmean = np.mean([x,y,z], axis=1)
    cmed = np.median([x,y,z], axis=1)

    ax.scatter(cmean[0], cmean[1], cmean[2], color='r', marker='s')

    ce90 = geolib.CE90(x,y)
    le90 = geolib.LE90(z)
    coefs = [ce90, ce90, le90]
    ax.set_title("CE90: %0.2f, LE90: %0.2f, n=%i" % (ce90, le90, x.shape[0])) 
    
    maxdim = np.ceil(np.max([np.max(np.abs([x, y, z])), ce90, le90]))
    ax.set_xlim(-maxdim, maxdim)
    ax.set_ylim(-maxdim, maxdim)
    ax.set_zlim(-maxdim, maxdim)

    rx, ry, rz = coefs
    u = np.linspace(0, 2 * np.pi, 100)
    v = np.linspace(0, np.pi, 100)
    ex = rx * np.outer(np.cos(u), np.sin(v))
    ey = ry * np.outer(np.sin(u), np.sin(v))
    ez = rz * np.outer(np.ones_like(u), np.cos(v))
    ax.plot_surface(ex, ey, ez, rstride=2, cstride=2, linewidth=0, color='b', alpha=0.1)
    #max_radius = max(rx, ry, rz)
    #for axis in 'xyz':
    #    getattr(ax, 'set_{}lim'.format(axis))((-max_radius, max_radius))
    if orthogonal_fig:
        from matplotlib.patches import Ellipse
        fig_ortho = plt.figure(figsize=(10,4))
        #fig_ortho = plt.figure()
        title='ICP Alignment Translation Vectors\nn=%i, mean: (%0.2f, %0.2f, %0.2f)\nCE90: %0.2f, LE90: %0.2f' % (x.shape[0], cmean[0], cmean[1], cmean[2], ce90, le90)
        plt.suptitle(title) 

        ax = fig_ortho.add_subplot(131)
        ax.plot(x, y, color='b', linestyle='None', marker='o', label='ICP correction vector')
        ax.plot(cmean[0], cmean[1], color='r', linestyle='None', marker='s', label='Mean')
        #ax.scatter(x, y)
        #ax.scatter(cmean[0], cmean[1], color='r', marker='s')
        ax.set_xlim(-maxdim, maxdim)
        ax.set_ylim(-maxdim, maxdim)
        ax.minorticks_on()
        ax.set_aspect('equal')
        ax.set_xlabel('X offset (m)')
        ax.set_ylabel('Y offset (m)')
        e = Ellipse((0,0), 2*ce90, 2*ce90, linewidth=0, alpha=0.1)
        ax.add_artist(e)
        plt.legend(prop={'size':8}, numpoints=1, loc='upper left')

        ax = fig_ortho.add_subplot(132)
        ax.plot(x, z, color='b', linestyle='None', marker='o', label='ICP correction vector')
        ax.plot(cmean[0], cmean[2], color='r', linestyle='None', marker='s', label='Mean')
        #ax.scatter(x, z)
        #ax.scatter(cmean[0], cmean[2], color='r', marker='s')
        ax.set_xlim(-maxdim, maxdim)
        ax.set_ylim(-maxdim, maxdim)
        ax.minorticks_on()
        ax.set_aspect('equal')
        ax.set_xlabel('X offset (m)')
        ax.set_ylabel('Z offset (m)')
        e = Ellipse((0,0), 2*ce90, 2*le90, linewidth=0, alpha=0.1)
        ax.add_artist(e)

        ax = fig_ortho.add_subplot(133)
        ax.plot(y, z, color='b', linestyle='None', marker='o', label='ICP correction vector')
        ax.plot(cmean[1], cmean[2], color='r', linestyle='None', marker='s', label='Mean')
        #ax.scatter(y, z)
        #ax.scatter(cmean[1], cmean[2], color='r', marker='s')
        ax.set_xlim(-maxdim, maxdim)
        ax.set_ylim(-maxdim, maxdim)
        ax.minorticks_on()
        ax.set_aspect('equal')
        ax.set_xlabel('Y offset (m)')
        ax.set_ylabel('Z offset (m)')
        e = Ellipse((0,0), 2*ce90, 2*le90, linewidth=0, alpha=0.1)
        ax.add_artist(e)
        
        plt.tight_layout()

        #Note: postscript doesn't properly handle tansparency
        #fig_fn = 'icp_translation_vec_proj_meters_orthogonal.pdf'
        fig_fn = 'icp_translation_vec_local_meters_orthogonal.pdf'
        plt.savefig(fig_fn, dpi=600, bbox_inches='tight')

    #Set back to original figure
    plt.figure(fig.number)

def main():
    #filenames = !ls *align/*reference-DEM.tif
    #run ~/src/demtools/error_analysis.py $filenames.s

    if len(sys.argv) < 1:
        sys.exit('No input files provided')

    fn_list = sys.argv[1:]
    n_samp = len(fn_list)

    error_dict_list = []
    for fn in fn_list:
        ed = parse_pc_align_log(fn)
        if 'Translation vector (North-East-Down, meters)' in ed.keys(): 
            error_dict_list.append(ed)  

    import matplotlib.dates as mdates
    #This is used for interactive display of x-value in plot window 
    date_str = '%Y-%m-%d %H:%M'
    date_fmt = mdates.DateFormatter(date_str)
    #ax.fmt_xdata = mdates.DateFormatter(date_fmt)
    months = mdates.MonthLocator() 
    months_int = mdates.MonthLocator(interval=6)  # every n months 
    years = mdates.YearLocator()   # every year
    yearsFmt = mdates.DateFormatter('%Y')
    #ax.xaxis.set_major_formatter(yearsFmt)
    #ax.xaxis.set_major_locator(months_int3)

    print
    print "n:", len(error_dict_list) 

    """
    #ECEF translations
    #key = 'Translation vector (ECEF meters)'
    key = 'Translation vector (Cartesian, meters)'
    #key = 'Translation vector (meters)'
    val = np.array([e[key] for e in error_dict_list])
    #make_plot3d(val[:,0], val[:,1], val[:,2], title=key)
    ce90 = geolib.CE90(val[:,0], val[:,1])
    le90 = geolib.LE90(val[:,2])
    print
    print key
    print "CE90:", ce90 
    print "LE90:", le90 
    print

    #Proj translation
    key = 'Translation vector (Proj meters)' 
    val = np.array([e[key] for e in error_dict_list])
    ce90 = geolib.CE90(val[:,0], val[:,1])
    le90 = geolib.LE90(val[:,2])
    print
    print key
    print "CE90:", ce90 
    print "LE90:", le90 
    print
    print 'Centroid (mean) of offsets (Proj meters): ', np.mean(val, axis=0) 
    print 'Centroid (median) of offsets (Proj meters): ', np.median(val, axis=0) 
    """

    #NOTE: changed default to N-E-D on 9/18/15
    #Can have significant differences for local proj vs. polar stereographic proj
    #Should regenerate all previous figures

    #Local translation on ellipsoid
    #This appears to be local stereographic projection on ellipsoid
    key = 'Translation vector (North-East-Down, meters)'
    val = np.array([e[key] for e in error_dict_list])

    #Reformat (n, e, +d) for (x, y, +z) coord sys
    val[:,[0,1]] = val[:,[1,0]]
    val[:,2] *= -1
    ce90 = geolib.CE90(val[:,0], val[:,1])
    le90 = geolib.LE90(val[:,2])
    print
    print key
    print "CE90:", ce90 
    print "LE90:", le90 
    print
    print 'Centroid (mean) of offsets (local ned meters): ', np.mean(val, axis=0) 
    print 'Centroid (median) of offsets (local ned meters): ', np.median(val, axis=0) 

    #Remove vertical bias
    remove_vertbias = False 
    if remove_vertbias:
        print "Removing vertical bias: %0.2f" % np.mean(val, axis=0)[2]
        val[:,2] -= np.mean(val, axis=0)[2]

    remove_outliers = False 
    #Flag outliers
    x_mag = val[:,0]
    y_mag = val[:,1]
    h_mag = np.sqrt(val[:,0]**2 + val[:,1]**2)
    v_mag = val[:,2]
    mag = np.sqrt(val[:,0]**2 + val[:,1]**2 + val[:,2]**2)
    abs_thresh = 10.0
    p = 98.0
    p_thresh = np.percentile(h_mag, p)
    #print "Outliers with horiz error >= %0.2f (%0.1f%%)" % (p_thresh, p)
    print "Outliers:" 
    #idx = (h_mag >= p_thresh).nonzero()[0]
    idx = (h_mag >= ce90).nonzero()[0]
    idx = np.unique(np.hstack([idx, ((np.abs(v_mag) >= le90).nonzero()[0])]))

    #Print all
    #idx = np.arange(h_mag.size)
    #idx_sort = np.argsort(mag[idx])
    #idx = idx[idx_sort]

    print 'name, m, h, v, x, y, z'
    for i in idx:
        print error_dict_list[i]['File'], mag[i], h_mag[i], v_mag[i], val[i,0:3]
        #Delete from list
        if remove_outliers:
            print "Removing from calculation"
            del error_dict_list[i]

    if remove_vertbias or remove_outliers:
        print
        print "Updated values"
        print key
        print "CE90:", geolib.CE90(val[:,0], val[:,1])
        print "LE90:", geolib.LE90(val[:,2])
        print
        print 'Centroid (mean) of offsets (local ned meters): ', np.mean(val, axis=0) 
        print 'Centroid (median) of offsets (local ned meters): ', np.median(val, axis=0) 

    #Extract dates
    date_vec = np.array([e['Date'] for e in error_dict_list])
    x = date_vec

    make_plot3d(val[:,0], val[:,1], val[:,2], title=key)
    #Note: there is a bug in pdf that displayes surface lines
    #fig_fn = 'icp_translation_vec_proj_meters.png'
    fig_fn = 'icp_translation_vec_local_meters.png'
    #plt.savefig(fig_fn, dpi=600, bbox_inches='tight')

    fig, ax = plt.subplots(1)
    key = 'Translation vector (lat,lon,z)'
    plt.title('ICP translation vector (lat,lon,z): Z component')
    val = np.array([e[key] for e in error_dict_list])
    y = val[:,2]
    make_plot(x,y,c='b',label=key, abs=False)
    fig.autofmt_xdate()
    ax.xaxis.set_minor_locator(months)
    #ax.xaxis.set_major_locator(months_int)
    #ax.xaxis.set_major_formatter(date_fmt)
    ax.fmt_xdata = date_fmt
    ax.set_ylabel('Z offset (m)')

    fig, ax = plt.subplots(1)
    key = 'Translation vector magnitude (meters)'
    plt.title('ICP Translation vector magnitude (meters)')
    y = np.array([e[key] for e in error_dict_list])
    make_plot(x,y,c='b',label=key, abs=True)
    fig.autofmt_xdate()
    ax.xaxis.set_minor_locator(months)
    #ax.xaxis.set_major_locator(months_int)
    #ax.xaxis.set_major_formatter(date_fmt)
    ax.fmt_xdata = date_fmt
    ax.set_ylabel('Offset (m)')

    fig, ax = plt.subplots(1)
    key = 'Number of errors'
    plt.title('Number of error samples')
    nerr = np.array([e[key] for e in error_dict_list])
    make_plot(x,nerr,c='b',label=key)
    fig.autofmt_xdate()
    ax.xaxis.set_minor_locator(months)
    #ax.xaxis.set_major_locator(months_int)
    #ax.xaxis.set_major_formatter(date_fmt)
    ax.fmt_xdata = date_fmt
    ax.set_ylabel('N samples')

    """
    fig, ax = plt.subplots(1)
    plt.title('ICP Standard Deviation')
    key = 'Input Std Error'
    in_std = np.array([e[key] for e in error_dict_list])
    make_plot(x,in_std,c='r',label=key)
    key = 'Output Std Error'
    out_std = np.array([e[key] for e in error_dict_list])
    make_plot(x,out_std,c='b',label=key)
    fig.autofmt_xdate()
    ax.xaxis.set_minor_locator(months)
    #ax.xaxis.set_major_locator(months_int)
    #ax.xaxis.set_major_formatter(date_fmt)
    ax.fmt_xdata = date_fmt
    plt.legend(scatterpoints=1)
    """

    fig, ax = plt.subplots(1)
    plt.title('ICP Mean Error')
    key = 'Input Mean Error'
    in_mean = np.array([e[key] for e in error_dict_list])
    #make_plot(x,in_mean,c='r',label=key,yerr=in_std)
    make_plot(x,in_mean,c='r',label=key, abs=True)
    key = 'Output Mean Error'
    out_mean = np.array([e[key] for e in error_dict_list])
    #make_plot(x,out_mean,c='b',label=key,yerr=out_std)
    make_plot(x,out_mean,c='b',label=key, abs=True)
    fig.autofmt_xdate()
    ax.xaxis.set_minor_locator(months)
    #ax.xaxis.set_major_locator(months_int)
    #ax.xaxis.set_major_formatter(date_fmt)
    ax.fmt_xdata = date_fmt
    ax.set_ylabel('Mean error (m)')
    plt.legend(scatterpoints=1, loc='upper left', prop={'size':8})

    fig, ax = plt.subplots(1)
    plt.title('ICP Median Error')
    key = 'Input 16th Percentile Error'
    in_16p = np.array([e[key] for e in error_dict_list])
    key = 'Input 84th Percentile Error'
    in_84p = np.array([e[key] for e in error_dict_list])
    key = 'Input Median Error'
    in_med = np.array([e[key] for e in error_dict_list])
    make_plot(x,in_med,c='r',label=key,yerr=[in_med - in_16p, in_84p - in_med], abs=True)
    key = 'Output 16th Percentile Error'
    out_16p = np.array([e[key] for e in error_dict_list])
    key = 'Output 84th Percentile Error'
    out_84p = np.array([e[key] for e in error_dict_list])
    key = 'Output Median Error'
    out_med = np.array([e[key] for e in error_dict_list])
    make_plot(x,out_med,c='b',label=key,yerr=[out_med - out_16p, out_84p - out_med], abs=True)
    fig.autofmt_xdate()
    ax.fmt_xdata = mdates.DateFormatter(date_fmt)
    ax.xaxis.set_minor_locator(months)
    #ax.xaxis.set_major_locator(months_int)
    #ax.xaxis.set_major_formatter(date_fmt)
    ax.fmt_xdata = date_fmt
    ax.set_ylabel('Median error (m)')
    plt.legend(scatterpoints=1, loc='upper left', prop={'size':8})
    fig_fn = 'icp_median_error.pdf'
    plt.savefig(fig_fn, dpi=600, bbox_inches='tight')

    fig, ax = plt.subplots(1)
    plt.title('Sampled Median Error')
    key = 'Input Sampled 16th Percentile Error'
    in_16p = np.ma.fix_invalid([e[key] for e in error_dict_list])
    if in_16p.count() > 0:
        key = 'Input Sampled 84th Percentile Error'
        in_84p = np.ma.fix_invalid([e[key] for e in error_dict_list])
        key = 'Input Sampled Median Error'
        in_med = np.ma.fix_invalid([e[key] for e in error_dict_list])
        in_spread = in_84p - in_16p
        make_plot(x,in_med,c='r',label=key,yerr=[in_med - in_16p, in_84p - in_med], abs=True)

        key = 'Output Sampled 16th Percentile Error'
        out_16p = np.ma.fix_invalid([e[key] for e in error_dict_list])
        key = 'Output Sampled 84th Percentile Error'
        out_84p = np.ma.fix_invalid([e[key] for e in error_dict_list])
        key = 'Output Sampled Median Error'
        out_med = np.ma.fix_invalid([e[key] for e in error_dict_list])
        out_spread = out_84p - out_16p

        p = 90.0
        out_med_thresh = np.percentile(out_med, p)
        out_spread_thresh = np.percentile(out_spread, p)
        #print "Outliers with horiz error >= %0.2f (%0.1f%%)" % (p_thresh, p)
        print
        print "Sampled Error Outliers:" 
        #idx = (h_mag >= p_thresh).nonzero()[0]
        idx = (out_med >= out_med_thresh).nonzero()[0]
        idx = np.unique(np.hstack([idx, ((out_spread >= out_spread_thresh).nonzero()[0])]))
        #Print all
        idx = np.arange(out_med.size)
        idx_sort = np.argsort(out_med[idx])
        idx = idx[idx_sort]
        print 'name, samp_mederrr, samp_errspread, nerr'
        for i in idx:
            print error_dict_list[i]['File'], out_med[i], out_spread[i], nerr[i]
            #Delete from list
            if remove_outliers:
                print "Removing from calculation"
                del error_dict_list[i]
        print
        print 'Input sampled median error (spread/2): %0.2f (%0.2f)' % (np.median(in_med), np.median(in_spread)/2.)
        print 'Output sampled median error (spread/2): %0.2f (%0.2f)' % (np.median(out_med), np.median(out_spread)/2.)
        print

        make_plot(x,out_med,c='b',label=key,yerr=[out_med - out_16p, out_84p - out_med], abs=True)
        fig.autofmt_xdate()
        ax.set_ylabel('Median error (m)')
        ax.fmt_xdata = mdates.DateFormatter(date_fmt)
        ax.xaxis.set_minor_locator(months)
        #ax.xaxis.set_major_locator(months_int)
        #ax.xaxis.set_major_formatter(date_fmt)
        ax.fmt_xdata = date_fmt
        ax.set_ylabel('Median error (m)')
        plt.legend(scatterpoints=1, loc='upper left', prop={'size':8})
        ax.set_ylim(-15,15)
        fig_fn = 'sampled_median_error.pdf'
        #fig_fn = 'sampled_median_error_2014-2016.pdf'
        #from datetime import datetime
        #ax.set_xlim(datetime(2014,1,1),datetime(2016,7,1))
        plt.savefig(fig_fn, dpi=600, bbox_inches='tight')
        #plt.show()

if __name__ == '__main__':
    main()



#! /usr/bin/env python

#Todo
#Better outlier removal
#Check Nuth and Kaab bin median
#Implement check for empty diff

import sys
import os
import argparse
import subprocess

from osgeo import gdal
import numpy as np
import matplotlib.pyplot as plt

from pygeotools.lib import iolib, malib, geolib, warplib, filtlib

from demcoreg import coreglib, dem_mask

from imview.lib import pltlib

#Turn off numpy multithreading
#os.environ['OPENBLAS_NUM_THREADS'] = '1'

def get_mask(ds, mask_list, dem_fn=None):
    #This returns True (1) for areas to mask, False (0) for valid static surfaces
    static_mask = dem_mask.get_mask(ds, mask_list, dem_fn, writeout=False)
    #return ~(static_mask)
    return static_mask

def outlier_filter(diff, f=3, perc=None, max_dz=100):
    print("Removing outliers")
    print("Initial pixel count:")
    print(diff.count())

    print("Absolute dz filter: %0.2f" % max_dz)
    #Absolute dz filter
    diff = np.ma.masked_greater(diff, max_dz)
    print(diff.count())

    if perc is not None:
        diff = filtlib.perc_fltr(diff, perc)
    else:
        #diff = filtlib.sigma_fltr(diff, f)
        diff = filtlib.mad_fltr(diff, f)

    print(diff.count())
    return diff

def get_filtered_slope(ds, slope_lim=(0.1, 40)):
    #Generate slope map
    print("Computing slope")
    slope = geolib.gdaldem_mem_ds(ds, processing='slope', returnma=True, computeEdges=False)
    #slope_stats = malib.print_stats(slope)
    print("Slope filter: %0.2f - %0.2f" % slope_lim)
    print("Initial count: %i" % slope.count()) 
    slope = filtlib.range_fltr(slope, slope_lim) 
    print(slope.count())
    return slope

def compute_offset(ref_dem_ds, src_dem_ds, src_dem_fn, mode='nuth', remove_outliers=True, max_offset=100, \
        max_dz=100, slope_lim=(0.1, 40), mask_list=['glaciers',], plot=True):
    #Make sure the input datasets have the same resolution/extent
    #Use projection of source DEM
    ref_dem_clip_ds, src_dem_clip_ds = warplib.memwarp_multi([ref_dem_ds, src_dem_ds], \
            res='max', extent='intersection', t_srs=src_dem_ds, r='cubic')

    #Compute size of NCC and SAD search window in pixels
    res = float(geolib.get_res(ref_dem_clip_ds, square=True)[0])
    max_offset_px = (max_offset/res) + 1
    #print(max_offset_px)
    pad = (int(max_offset_px), int(max_offset_px))

    #This will be updated geotransform for src_dem
    src_dem_gt = np.array(src_dem_clip_ds.GetGeoTransform())

    #Load the arrays
    ref_dem = iolib.ds_getma(ref_dem_clip_ds, 1)
    src_dem = iolib.ds_getma(src_dem_clip_ds, 1)

    print("Elevation difference stats for uncorrected input DEMs (src - ref)")
    diff = src_dem - ref_dem

    static_mask = get_mask(src_dem_clip_ds, mask_list, src_dem_fn)
    diff = np.ma.array(diff, mask=static_mask)

    if diff.count() == 0:
        sys.exit("No overlapping, unmasked pixels shared between input DEMs")

    if remove_outliers:
        diff = outlier_filter(diff, f=3, max_dz=max_dz)

    #Want to use higher quality DEM, should determine automatically from original res/count
    #slope = get_filtered_slope(ref_dem_clip_ds, slope_lim=slope_lim)
    slope = get_filtered_slope(src_dem_clip_ds, slope_lim=slope_lim)

    print("Computing aspect")
    #aspect = geolib.gdaldem_mem_ds(ref_dem_clip_ds, processing='aspect', returnma=True, computeEdges=False)
    aspect = geolib.gdaldem_mem_ds(src_dem_clip_ds, processing='aspect', returnma=True, computeEdges=False)

    ref_dem_clip_ds = None
    src_dem_clip_ds = None

    #Apply slope filter to diff
    #Note that we combine masks from diff and slope in coreglib
    diff = np.ma.array(diff, mask=np.ma.getmaskarray(slope))

    #Get final mask after filtering
    static_mask = np.ma.getmaskarray(diff)

    #Compute stats for new masked difference map
    print("Filtered difference map")
    diff_stats = malib.print_stats(diff)
    dz = diff_stats[5]

    print("Computing sub-pixel offset between DEMs using mode: %s" % mode)

    #By default, don't create output figure
    fig = None

    #Default horizntal shift is (0,0)
    dx = 0
    dy = 0

    #Sum of absolute differences
    if mode == "sad":
        ref_dem = np.ma.array(ref_dem, mask=static_mask)
        src_dem = np.ma.array(src_dem, mask=static_mask)
        m, int_offset, sp_offset = coreglib.compute_offset_sad(ref_dem, src_dem, pad=pad)
        #Geotransform has negative y resolution, so don't need negative sign
        #np array is positive down
        #GDAL coordinates are positive up
        dx = sp_offset[1]*src_dem_gt[1]
        dy = sp_offset[0]*src_dem_gt[5]
    #Normalized cross-correlation of clipped, overlapping areas
    elif mode == "ncc":
        ref_dem = np.ma.array(ref_dem, mask=static_mask)
        src_dem = np.ma.array(src_dem, mask=static_mask)
        m, int_offset, sp_offset, fig = coreglib.compute_offset_ncc(ref_dem, src_dem, \
                pad=pad, prefilter=False, plot=plot)
        dx = sp_offset[1]*src_dem_gt[1]
        dy = sp_offset[0]*src_dem_gt[5]
    #Nuth and Kaab (2011)
    elif mode == "nuth":
        #Compute relationship between elevation difference, slope and aspect
        fit_param, fig = coreglib.compute_offset_nuth(diff, slope, aspect, plot=plot)
        if fit_param is None:
            print("Failed to calculate horizontal shift")
        else:
            #fit_param[0] is magnitude of shift vector
            #fit_param[1] is direction of shift vector
            #fit_param[2] is mean bias divided by tangent of mean slope
            #print(fit_param)
            dx = fit_param[0]*np.sin(np.deg2rad(fit_param[1]))
            dy = fit_param[0]*np.cos(np.deg2rad(fit_param[1]))
            med_slope = malib.fast_median(slope)
            nuth_dz = fit_param[2]*np.tan(np.deg2rad(med_slope))
            print('Median dz: %0.2f\nNuth dz: %0.2f' % (dz, nuth_dz))
            #dz = nuth_dz
    elif mode == "all":
        print("Not yet implemented")
        #Want to compare all methods, average offsets
        #m, int_offset, sp_offset = coreglib.compute_offset_sad(ref_dem, src_dem)
        #m, int_offset, sp_offset = coreglib.compute_offset_ncc(ref_dem, src_dem)
    elif mode == "none":
        print("Skipping alignment, writing out DEM with median bias over static surfaces removed")
        dst_fn = outprefix+'_med%0.1f.tif' % dz
        iolib.writeGTiff(src_dem_orig + dz, dst_fn, src_dem_ds)
        sys.exit()
    #Note: minus signs here since we are computing dz=(src-ref), but adjusting src
    return -dx, -dy, -dz, static_mask, fig

def getparser():
    parser = argparse.ArgumentParser(description="Perform DEM co-registration using multiple algorithms", \
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('ref_fn', type=str, help='Reference DEM filename')
    parser.add_argument('src_fn', type=str, help='Source DEM filename to be shifted')
    parser.add_argument('-mode', type=str, default='nuth', choices=['ncc', 'sad', 'nuth', 'none'], \
            help='Type of co-registration to use')
    parser.add_argument('-mask_list', nargs='+', type=str, default=['glaciers',], choices=dem_mask.mask_choices, \
            help='Define masks to use to limit reference surfaces for co-registration')
    parser.add_argument('-tiltcorr', action='store_true', \
            help='After preliminary translation, fit 2D polynomial to residual elevation offsets and remove')
    parser.add_argument('-polyorder', type=int, default=1, \
            help='Specify order of 2D polynomial fit') 
    parser.add_argument('-tol', type=float, default=0.02, \
            help='When iterative translation magnitude is below this tolerance (meters), break and write out corrected DEM')
    parser.add_argument('-max_offset', type=float, default=100, \
            help='Maximum expected horizontal offset in meters, used to set search range for ncc and sad modes')
    parser.add_argument('-max_dz', type=float, default=100, \
            help='Maximum expected vertical offset in meters, used to filter outliers')
    res_choices = ['min', 'max', 'mean', 'common_scale_factor']
    parser.add_argument('-res', type=str, default='mean', choices=res_choices, \
            help='Warp intputs to this resolution') 
    parser.add_argument('-slope_lim', type=float, nargs=2, default=(0.1, 40), \
            help='Minimum and maximum surface slope limits to consider')
    parser.add_argument('-max_iter', type=int, default=30, \
            help='Maximum number of iterations, if tol is not reached')
    parser.add_argument('-outdir', default=None, help='Output directory')
    return parser

def main(argv=None):
    parser = getparser()
    args = parser.parse_args()

    #Should check that files exist
    ref_dem_fn = args.ref_fn
    src_dem_fn = args.src_fn

    mode = args.mode
    mask_list = args.mask_list
    max_offset = args.max_offset
    max_dz = args.max_dz
    slope_lim = tuple(args.slope_lim)
    tiltcorr = args.tiltcorr
    polyorder = args.polyorder
    res = args.res

    #Maximum number of iterations
    max_iter = args.max_iter

    #These are tolerances (in meters) to stop iteration
    tol = args.tol
    min_dx = tol
    min_dy = tol
    min_dz = tol

    outdir = args.outdir
    if outdir is None:
        outdir = os.path.splitext(src_dem_fn)[0] + '_dem_align'

    if tiltcorr:
        outdir += '_tiltcorr'
        tiltcorr_done = False
        #Relax tolerance for initial round of co-registration
        #tiltcorr_tol = 0.1
        #if tol < tiltcorr_tol:
        #    tol = tiltcorr_tol

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    outprefix = '%s_%s' % (os.path.splitext(os.path.split(src_dem_fn)[-1])[0], \
            os.path.splitext(os.path.split(ref_dem_fn)[-1])[0])
    outprefix = os.path.join(outdir, outprefix)

    print("\nReference: %s" % ref_dem_fn)
    print("Source: %s" % src_dem_fn)
    print("Mode: %s" % mode)
    print("Output: %s\n" % outprefix)

    src_dem_ds = gdal.Open(src_dem_fn)
    ref_dem_ds = gdal.Open(ref_dem_fn)

    #Get local cartesian coordinate system
    #local_srs = geolib.localtmerc_ds(src_dem_ds)
    #Use original source dataset coordinate system
    #Potentially issues with distortion and xyz/tiltcorr offsets for DEM with large extent
    local_srs = geolib.get_ds_srs(src_dem_ds)
    #local_srs = geolib.get_ds_srs(ref_dem_ds)

    #Resample to common grid
    ref_dem_res = float(geolib.get_res(ref_dem_ds, t_srs=local_srs, square=True)[0])
    #Create a copy to be updated in place
    src_dem_ds_align = iolib.mem_drv.CreateCopy('', src_dem_ds, 0)
    src_dem_res = float(geolib.get_res(src_dem_ds, t_srs=local_srs, square=True)[0])
    src_dem_ds = None
    #Resample to user-specified resolution
    ref_dem_ds, src_dem_ds_align = warplib.memwarp_multi([ref_dem_ds, src_dem_ds_align], \
            extent='intersection', res=args.res, t_srs=local_srs, r='cubic')

    res = float(geolib.get_res(src_dem_ds_align, square=True)[0])
    print("\nReference DEM res: %0.2f" % ref_dem_res)
    print("Source DEM res: %0.2f" % src_dem_res)
    print("Resolution for coreg: %s (%0.2f m)\n" % (args.res, res))

    #Iteration number
    n = 1
    #Cumulative offsets
    dx_total = 0
    dy_total = 0
    dz_total = 0

    #Now iteratively update geotransform and vertical shift
    while True:
        print("*** Iteration %i ***" % n)
        dx, dy, dz, static_mask, fig = compute_offset(ref_dem_ds, src_dem_ds_align, src_dem_fn, mode, max_offset, \
                mask_list=mask_list, max_dz=max_dz, slope_lim=slope_lim, plot=True)
        xyz_shift_str_iter = "dx=%+0.2fm, dy=%+0.2fm, dz=%+0.2fm" % (dx, dy, dz)
        print("Incremental offset: %s" % xyz_shift_str_iter)

        dx_total += dx
        dy_total += dy
        dz_total += dz

        xyz_shift_str_cum = "dx=%+0.2fm, dy=%+0.2fm, dz=%+0.2fm" % (dx_total, dy_total, dz_total)
        print("Cumulative offset: %s" % xyz_shift_str_cum)
        #String to append to output filenames
        xyz_shift_str_cum_fn = '_%s_x%+0.2f_y%+0.2f_z%+0.2f' % (mode, dx_total, dy_total, dz_total)

        #Should make an animation of this converging
        if n == 1: 
            #static_mask_orig = static_mask
            if fig is not None:
                dst_fn = outprefix + '_%s_iter%02i_plot.png' % (mode, n)
                print("Writing offset plot: %s" % dst_fn)
                fig.gca().set_title("Incremental: %s\nCumulative: %s" % (xyz_shift_str_iter, xyz_shift_str_cum))
                fig.savefig(dst_fn, dpi=300)

        #Apply the horizontal shift to the original dataset
        src_dem_ds_align = coreglib.apply_xy_shift(src_dem_ds_align, dx, dy, createcopy=False)
        #Should 
        src_dem_ds_align = coreglib.apply_z_shift(src_dem_ds_align, dz, createcopy=False)

        n += 1
        print("\n")
        #If magnitude of shift in all directions is less than tol
        #if n > max_iter or (abs(dx) <= min_dx and abs(dy) <= min_dy and abs(dz) <= min_dz):
        #If magnitude of shift is less than tol
        dm = np.sqrt(dx**2 + dy**2 + dz**2)
        dm_total = np.sqrt(dx_total**2 + dy_total**2 + dz_total**2)

        if dm_total > max_offset:
            sys.exit("Total offset exceeded specified max_offset (%0.2f m). Consider increasing -max_offset argument" % max_offset)

        #Stop iteration
        if n > max_iter or dm < tol:

            if fig is not None:
                dst_fn = outprefix + '_%s_iter%02i_plot.png' % (mode, n)
                print("Writing offset plot: %s" % dst_fn)
                fig.gca().set_title("Incremental:%s\nCumulative:%s" % (xyz_shift_str_iter, xyz_shift_str_cum))
                fig.savefig(dst_fn, dpi=300)

            #Compute final elevation difference
            if True:
                ref_dem_clip_ds_align, src_dem_clip_ds_align = warplib.memwarp_multi([ref_dem_ds, src_dem_ds_align], \
                        res=res, extent='intersection', t_srs=local_srs, r='cubic')
                ref_dem_align = iolib.ds_getma(ref_dem_clip_ds_align, 1)
                src_dem_align = iolib.ds_getma(src_dem_clip_ds_align, 1)
                ref_dem_clip_ds_align = None

                diff_align = src_dem_align - ref_dem_align
                src_dem_align = None
                ref_dem_align = None

                #Get updated, final mask
                static_mask_final = get_mask(src_dem_clip_ds_align, mask_list, src_dem_fn)
                static_mask_final = np.logical_or(np.ma.getmaskarray(diff_align), static_mask_final)
                
                #Final stats, before outlier removal
                diff_align_compressed = diff_align[~static_mask_final]
                diff_align_stats = malib.get_stats_dict(diff_align_compressed, full=True)

                #Prepare filtered version for tiltcorr fit
                diff_align_filt = np.ma.array(diff_align, mask=static_mask_final)
                diff_align_filt = outlier_filter(diff_align_filt, f=3, max_dz=max_dz)
                #diff_align_filt = outlier_filter(diff_align_filt, perc=(12.5, 87.5), max_dz=max_dz)
                slope = get_filtered_slope(src_dem_clip_ds_align)
                diff_align_filt = np.ma.array(diff_align_filt, mask=np.ma.getmaskarray(slope))
                diff_align_filt_stats = malib.get_stats_dict(diff_align_filt, full=True)

            #Fit 2D polynomial to residuals and remove
            #To do: add support for along-track and cross-track artifacts
            if tiltcorr and not tiltcorr_done:
                print("\n************")
                print("Calculating 'tiltcorr' 2D polynomial fit to residuals with order %i" % polyorder)
                print("************\n")
                gt = src_dem_clip_ds_align.GetGeoTransform()

                #Need to apply the mask here, so we're only fitting over static surfaces
                #Note that the origmask=False will compute vals for all x and y indices, which is what we want
                vals, resid, coeff = geolib.ma_fitpoly(diff_align_filt, order=polyorder, gt=gt, perc=(0,100), origmask=False)
                #vals, resid, coeff = geolib.ma_fitplane(diff_align_filt, gt, perc=(12.5, 87.5), origmask=False)

                #Should write out coeff or grid with correction 

                vals_stats = malib.get_stats_dict(vals)

                #Want to have max_tilt check here
                #max_tilt = 4.0 #m
                #Should do percentage
                #vals.ptp() > max_tilt

                #Note: dimensions of ds and vals will be different as vals are computed for clipped intersection
                #Need to recompute planar offset for full src_dem_ds_align extent and apply
                xgrid, ygrid = geolib.get_xy_grids(src_dem_ds_align)
                valgrid = geolib.polyval2d(xgrid, ygrid, coeff) 
                #For results of ma_fitplane
                #valgrid = coeff[0]*xgrid + coeff[1]*ygrid + coeff[2]
                src_dem_ds_align = coreglib.apply_z_shift(src_dem_ds_align, -valgrid, createcopy=False)

                if True:
                    print("Creating plot of polynomial fit to residuals")
                    fig, axa = plt.subplots(1,2, figsize=(8, 4))
                    dz_clim = malib.calcperc_sym(vals, (2, 98))
                    ax = pltlib.iv(diff_align_filt, ax=axa[0], cmap='RdBu', clim=dz_clim, \
                            label='Residual dz (m)', scalebar=False)
                    ax = pltlib.iv(valgrid, ax=axa[1], cmap='RdBu', clim=dz_clim, \
                            label='Polyfit dz (m)', ds=src_dem_ds_align)
                    #if tiltcorr:
                        #xyz_shift_str_cum_fn += "_tiltcorr"
                    tiltcorr_fig_fn = outprefix + '%s_polyfit.png' % xyz_shift_str_cum_fn
                    print("Writing out figure: %s\n" % tiltcorr_fig_fn)
                    fig.savefig(tiltcorr_fig_fn, dpi=300)

                print("Applying tilt correction to difference map")
                diff_align -= vals

                #Should iterate until tilts are below some threshold
                #For now, only do one tiltcorr
                tiltcorr_done=True
                #Now use original tolerance, and number of iterations 
                tol = args.tol
                max_iter = n + args.max_iter
            else:
                break

    if True:
        #Write out aligned difference map for clipped extent with vertial offset removed
        align_diff_fn = outprefix + '%s_align_diff.tif' % xyz_shift_str_cum_fn
        print("Writing out aligned difference map with median vertical offset removed")
        iolib.writeGTiff(diff_align, align_diff_fn, src_dem_clip_ds_align)

    if True:
        #Write out fitered aligned difference map
        align_diff_filt_fn = outprefix + '%s_align_diff_filt.tif' % xyz_shift_str_cum_fn
        print("Writing out filtered aligned difference map with median vertical offset removed")
        iolib.writeGTiff(diff_align_filt, align_diff_filt_fn, src_dem_clip_ds_align)

    #Extract final center coordinates for intersection
    center_coord_ll = geolib.get_center(src_dem_clip_ds_align, t_srs=geolib.wgs_srs)
    center_coord_xy = geolib.get_center(src_dem_clip_ds_align)
    src_dem_clip_ds_align = None

    #Write out final aligned src_dem 
    align_fn = outprefix + '%s_align.tif' % xyz_shift_str_cum_fn
    print("Writing out shifted src_dem with median vertical offset removed: %s" % align_fn)
    #Open original uncorrected dataset at native resolution
    src_dem_ds = gdal.Open(src_dem_fn)
    src_dem_ds_align = iolib.mem_drv.CreateCopy('', src_dem_ds, 0)
    #Apply final horizontal and vertial shift to the original dataset
    #Note: potentially issues if we used a different projection during coregistration!
    src_dem_ds_align = coreglib.apply_xy_shift(src_dem_ds_align, dx_total, dy_total, createcopy=False)
    src_dem_ds_align = coreglib.apply_z_shift(src_dem_ds_align, dz_total, createcopy=False)
    if tiltcorr:
        xgrid, ygrid = geolib.get_xy_grids(src_dem_ds_align)
        valgrid = geolib.polyval2d(xgrid, ygrid, coeff) 
        #For results of ma_fitplane
        #valgrid = coeff[0]*xgrid + coeff[1]*ygrid + coeff[2]
        src_dem_ds_align = coreglib.apply_z_shift(src_dem_ds_align, -valgrid, createcopy=False)
    #Might be cleaner way to write out MEM ds directly to disk
    src_dem_full_align = iolib.ds_getma(src_dem_ds_align)
    iolib.writeGTiff(src_dem_full_align, align_fn, src_dem_ds_align)

    if True:
        #Output final aligned src_dem, masked so only best pixels are preserved
        #Useful if creating a new reference product
        #Can also use apply_mask.py 
        print("Applying filter to shiftec src_dem")
        align_diff_filt_full_ds = warplib.memwarp_multi_fn([align_diff_filt_fn,], res=src_dem_ds_align, extent=src_dem_ds_align, \
                t_srs=src_dem_ds_align)[0]
        align_diff_filt_full = iolib.ds_getma(align_diff_filt_full_ds)
        align_diff_filt_full_ds = None
        align_fn_masked = outprefix + '%s_align_filt.tif' % xyz_shift_str_cum_fn
        iolib.writeGTiff(np.ma.array(src_dem_full_align, mask=np.ma.getmaskarray(align_diff_filt_full)), \
                align_fn_masked, src_dem_ds_align)

    src_dem_full_align = None
    src_dem_ds_align = None

    #Compute original elevation difference
    if True:
        ref_dem_clip_ds, src_dem_clip_ds = warplib.memwarp_multi([ref_dem_ds, src_dem_ds], \
                res=res, extent='intersection', t_srs=local_srs, r='cubic')
        src_dem_ds = None
        ref_dem_ds = None
        ref_dem_orig = iolib.ds_getma(ref_dem_clip_ds)
        src_dem_orig = iolib.ds_getma(src_dem_clip_ds)
        #Needed for plotting
        ref_dem_hs = geolib.gdaldem_mem_ds(ref_dem_clip_ds, processing='hillshade', returnma=True, computeEdges=True)
        src_dem_hs = geolib.gdaldem_mem_ds(src_dem_clip_ds, processing='hillshade', returnma=True, computeEdges=True)
        diff_orig = src_dem_orig - ref_dem_orig
        #Only compute stats over valid surfaces
        static_mask_orig = get_mask(src_dem_clip_ds, mask_list, src_dem_fn)
        #Note: this doesn't include outlier removal or slope mask!
        static_mask_orig = np.logical_or(np.ma.getmaskarray(diff_orig), static_mask_orig)
        #For some reason, ASTER DEM diff have a spike near the 0 bin, could be an issue with masking?
        diff_orig_compressed = diff_orig[~static_mask_orig]
        diff_orig_stats = malib.get_stats_dict(diff_orig_compressed, full=True)

        #Prepare filtered version for comparison 
        diff_orig_filt = np.ma.array(diff_orig, mask=static_mask_orig)
        diff_orig_filt = outlier_filter(diff_orig_filt, f=3, max_dz=max_dz)
        #diff_orig_filt = outlier_filter(diff_orig_filt, perc=(12.5, 87.5), max_dz=max_dz)
        slope = get_filtered_slope(src_dem_clip_ds)
        diff_orig_filt = np.ma.array(diff_orig_filt, mask=np.ma.getmaskarray(slope))
        diff_orig_filt_stats = malib.get_stats_dict(diff_orig_filt, full=True)

        #Write out original difference map
        print("Writing out original difference map for common intersection before alignment")
        orig_diff_fn = outprefix + '_orig_diff.tif'
        iolib.writeGTiff(diff_orig, orig_diff_fn, ref_dem_clip_ds)
        src_dem_clip_ds = None
        ref_dem_clip_ds = None

    if True:
        align_stats_fn = outprefix + '%s_align_stats.json' % xyz_shift_str_cum_fn
        align_stats = {}
        align_stats['src_fn'] = src_dem_fn 
        align_stats['ref_fn'] = ref_dem_fn 
        align_stats['align_fn'] = align_fn 
        align_stats['res'] = {} 
        align_stats['res']['src'] = src_dem_res
        align_stats['res']['ref'] = ref_dem_res
        align_stats['res']['coreg'] = res
        align_stats['center_coord'] = {'lon':center_coord_ll[0], 'lat':center_coord_ll[1], \
                'x':center_coord_xy[0], 'y':center_coord_xy[1]}
        align_stats['shift'] = {'dx':dx_total, 'dy':dy_total, 'dz':dz_total, 'dm':dm_total}
        #This tiltcorr flag gets set to false, need better flag
        if tiltcorr:
            align_stats['tiltcorr'] = {}
            align_stats['tiltcorr']['coeff'] = coeff.tolist()
            align_stats['tiltcorr']['val_stats'] = vals_stats
        align_stats['before'] = diff_orig_stats
        align_stats['before_filt'] = diff_orig_filt_stats
        align_stats['after'] = diff_align_stats
        align_stats['after_filt'] = diff_align_filt_stats
        
        import json
        with open(align_stats_fn, 'w') as f:
            json.dump(align_stats, f)

    #Create output plot
    if True:
        print("Creating final plot")
        #f, axa = plt.subplots(2, 4, figsize=(11, 8.5))
        f, axa = plt.subplots(2, 4, figsize=(16, 8))
        for ax in axa.ravel()[:-1]:
            ax.set_facecolor('k')
            pltlib.hide_ticks(ax)
        dem_clim = malib.calcperc(ref_dem_orig, (2,98))
        axa[0,0].imshow(ref_dem_hs, cmap='gray')
        im = axa[0,0].imshow(ref_dem_orig, cmap='cpt_rainbow', clim=dem_clim, alpha=0.6)
        pltlib.add_cbar(axa[0,0], im, arr=ref_dem_orig, clim=dem_clim, label=None)
        pltlib.add_scalebar(axa[0,0], res=res)
        axa[0,0].set_title('Reference DEM')
        axa[0,1].imshow(src_dem_hs, cmap='gray')
        im = axa[0,1].imshow(src_dem_orig, cmap='cpt_rainbow', clim=dem_clim, alpha=0.6)
        pltlib.add_cbar(axa[0,1], im, arr=src_dem_orig, clim=dem_clim, label=None)
        axa[0,1].set_title('Source DEM')
        #axa[0,2].imshow(~static_mask_orig, clim=(0,1), cmap='gray')
        axa[0,2].imshow(~static_mask, clim=(0,1), cmap='gray')
        axa[0,2].set_title('Surfaces for co-registration')
        dz_clim = malib.calcperc_sym(diff_orig_compressed, (5, 95))
        im = axa[1,0].imshow(diff_orig, cmap='RdBu', clim=dz_clim)
        pltlib.add_cbar(axa[1,0], im, arr=diff_orig, clim=dz_clim, label=None)
        axa[1,0].set_title('Elev. Diff. Before (m)')
        im = axa[1,1].imshow(diff_align, cmap='RdBu', clim=dz_clim)
        pltlib.add_cbar(axa[1,1], im, arr=diff_align, clim=dz_clim, label=None)
        axa[1,1].set_title('Elev. Diff. After (m)')

        #tight_dz_clim = (-1.0, 1.0)
        tight_dz_clim = (-2.0, 2.0)
        #tight_dz_clim = (-10.0, 10.0)
        #tight_dz_clim = malib.calcperc_sym(diff_align_filt, (5, 95))
        im = axa[1,2].imshow(diff_align_filt, cmap='RdBu', clim=tight_dz_clim)
        pltlib.add_cbar(axa[1,2], im, arr=diff_align_filt, clim=tight_dz_clim, label=None)
        axa[1,2].set_title('Elev. Diff. After (m)')

        #Tried to insert Nuth fig here
        #ax_nuth.change_geometry(1,2,1)
        #f.axes.append(ax_nuth)

        bins = np.linspace(dz_clim[0], dz_clim[1], 128)
        axa[1,3].hist(diff_orig_compressed, bins, color='g', label='Before', alpha=0.5)
        axa[1,3].hist(diff_align_compressed, bins, color='b', label='After', alpha=0.5)
        axa[1,3].set_xlim(*dz_clim)
        axa[1,3].axvline(0, color='k', linewidth=0.5, linestyle=':')
        axa[1,3].set_xlabel('Elev. Diff. (m)')
        axa[1,3].set_ylabel('Count (px)')
        axa[1,3].set_title("Source - Reference")
        before_str = 'Before\nmed: %0.2f\nnmad: %0.2f' % (diff_orig_stats['med'], diff_orig_stats['nmad'])
        axa[1,3].text(0.05, 0.95, before_str, va='top', color='g', transform=axa[1,3].transAxes, fontsize=8)
        after_str = 'After\nmed: %0.2f\nnmad: %0.2f' % (diff_align_stats['med'], diff_align_stats['nmad'])
        axa[1,3].text(0.65, 0.95, after_str, va='top', color='b', transform=axa[1,3].transAxes, fontsize=8)

        #This is empty
        axa[0,3].axis('off')

        suptitle = '%s\nx: %+0.2fm, y: %+0.2fm, z: %+0.2fm' % (os.path.split(outprefix)[-1], dx_total, dy_total, dz_total)
        f.suptitle(suptitle)
        f.tight_layout()
        plt.subplots_adjust(top=0.90)

        fig_fn = outprefix + '%s_align.png' % xyz_shift_str_cum_fn
        print("Writing out figure: %s" % fig_fn)
        f.savefig(fig_fn, dpi=300)

if __name__ == "__main__":
    main()



#! /usr/bin/env python

#Filter preprocessed ICESat-1 GLAS points for a given input raster

#First run glas_proc.py - see usage
#Also expects *DEM_32m_ref.tif output from dem_mask.py

#Then run filter_glas.py for a single DEM fn or a list of DEM fn
#parallel --jobs 16 --delay 1 '~/src/demcoreg/demcoreg/filter_glas.py {}' ::: */dem*/*DEM_32m.tif

import sys
import os

import numpy as np
from osgeo import gdal
from pygeotools.lib import geolib, iolib, malib, timelib

import matplotlib.pyplot as plt

from imview.lib import gmtColormap, pltlib
cpt_rainbow = gmtColormap.get_rainbow()

site = 'hma'

#Minimum number of points required to write out _ref.csv
min_pts = 100

#Maximum value of surface slope to use
max_slope = 20.

pt_srs = geolib.wgs_srs
#This is time column in YYYYMMDD
tcol = 0
xcol = 3
ycol = 2
zcol = 4

#Padding in pixels for sample radius
#Since we're likely dealing with 32-m products here, can just use pad=1
pad = 1
#pad = 'glas'

glas_dir = '/nobackupp8/deshean/icesat_glas'
#ext = 'GLAH14_%s_refdemfilt_lulcfilt' % site
ext = 'GLAH14_%s_refdemfilt' % site
glas_npz_fn = os.path.join(glas_dir, ext+'.npz')

if not os.path.exists(glas_npz_fn):
    glas_csv_fn = os.path.splitext(glas_npz_fn)[0]+'.csv'
    print("Loading csv: %s" % glas_csv_fn)
    glas_pts = np.loadtxt(glas_csv_fn, delimiter=',', skiprows=1, dtype=None)
    print("Saving npz: %s" % glas_npz_fn)
    np.savez_compressed(glas_npz_fn, glas_pts)
else:
    #This takes ~5 seconds to load ~9M records with 8 fields
    print("Loading npz: %s" % glas_npz_fn)
    glas_pts = np.load(glas_npz_fn)['arr_0']

dem_fn_list = sys.argv[1:]
for n,dem_fn in enumerate(dem_fn_list):
    print("%i of %i" % (n+1, len(dem_fn_list)))
    #Lat/lon extent filter
    print("Loading DEM: %s" % dem_fn)
    dem_ds = gdal.Open(dem_fn)
    dem_ma = iolib.ds_getma(dem_ds)
    dem_extent_wgs84 = geolib.ds_extent(dem_ds, t_srs=pt_srs)
    xmin, ymin, xmax, ymax = dem_extent_wgs84
    print("Applying spatial filter") 
    x = glas_pts[:,xcol]
    y = glas_pts[:,ycol]
    idx = ((x >= xmin) & (x <= xmax) & (y >= ymin) & (y <= ymax)) 
    if idx.nonzero()[0].size == 0:
        print("No points after spatial filter")
        continue

    print("Sampling DEM at masked point locations") 
    glas_pts_fltr = glas_pts[idx]

    print("Writing out %i points after spatial filter" % glas_pts_fltr.shape[0]) 
    out_csv_fn = os.path.splitext(dem_fn)[0]+'_%s.csv' % ext

    # dt_ordinal, dt_YYYYMMDD, lat, lon, z_WGS84 
    fmt = '%0.8f, %i, %0.6f, %0.6f, %0.2f'
    if glas_pts_fltr.shape[1] == 7:
        # dt_ordinal, dt_YYYYMMDD, lat, lon, z_WGS84, z_refdem_med_WGS84, z_refdem_nmad
        fmt += ', %0.2f, %0.2f'
    elif glas_pts_fltr.shape[1] == 8:
        # dt_ordinal, dt_YYYYMMDD, lat, lon, z_WGS84, z_refdem_med_WGS84, z_refdem_nmad, lulc
        fmt += ', %0.2f, %0.2f, %i'
    np.savetxt(out_csv_fn, glas_pts_fltr, fmt=fmt, delimiter=',')

    x_fltr = glas_pts_fltr[:,xcol]
    y_fltr = glas_pts_fltr[:,ycol]
    z_fltr = glas_pts_fltr[:,zcol]

    dem_mask_fn = os.path.splitext(dem_fn)[0]+'_ref.tif'
    if os.path.exists(dem_mask_fn):
        print("Loading Masked DEM: %s" % dem_mask_fn)
        dem_mask_ds = gdal.Open(dem_mask_fn) 
        dem_mask = iolib.ds_getma(dem_mask_ds) 
    else:
        dem_mask_ds = dem_ds
        dem_mask = dem_ma

    #Convert input xy coordinates to raster coordinates
    mX_fltr, mY_fltr, mZ = geolib.cT_helper(x_fltr, y_fltr, 0, pt_srs, geolib.get_ds_srs(dem_mask_ds))
    pX_fltr, pY_fltr = geolib.mapToPixel(mX_fltr, mY_fltr, dem_mask_ds.GetGeoTransform())
    pX_fltr = np.atleast_1d(pX_fltr)
    pY_fltr = np.atleast_1d(pY_fltr)

    #Sample raster
    #This returns median and mad for ICESat footprint
    samp = geolib.sample(dem_mask_ds, mX_fltr, mY_fltr, pad=pad)
    samp_idx = ~(np.ma.getmaskarray(samp[:,0]))
    npts = samp_idx.nonzero()[0].size
    if npts < min_pts:
        print("Not enough points after sampling valud pixels, post bareground mask (%i < %i)" % (npts, min_pts))
        continue
       
    if True:
        print("Applying slope filter, masking points with slope > %0.1f" % max_slope)
        slope_ds = geolib.gdaldem_mem_ds(dem_mask_ds, processing='slope', returnma=False)
        slope_samp = geolib.sample(slope_ds, mX_fltr, mY_fltr, pad=pad)
        slope_samp_idx = (slope_samp[:,0] <= max_slope).data
        samp_idx = np.logical_and(slope_samp_idx, samp_idx)

    npts = samp_idx.nonzero()[0].size
    if npts < min_pts:
        print("Not enough points after %0.1f deg slope mask (%i < %i)" % (max_slope, npts, min_pts))
        continue

    glas_pts_fltr_mask = glas_pts_fltr[samp_idx]

    if os.path.exists(dem_mask_fn):
        print("Writing out %i points after mask" % glas_pts_fltr_mask.shape[0]) 
        out_csv_fn_mask = os.path.splitext(out_csv_fn)[0]+'_ref.csv'
        #Could add DEM samp columns here
        np.savetxt(out_csv_fn_mask, glas_pts_fltr_mask, fmt=fmt, delimiter=',')

    x_fltr_mask = glas_pts_fltr_mask[:,xcol]
    y_fltr_mask = glas_pts_fltr_mask[:,ycol]
    z_fltr_mask = glas_pts_fltr_mask[:,zcol]
    mX_fltr_mask, mY_fltr_mask, mZ = geolib.cT_helper(x_fltr_mask, y_fltr_mask, 0, pt_srs, geolib.get_ds_srs(dem_mask_ds))
    pX_fltr_mask, pY_fltr_mask = geolib.mapToPixel(mX_fltr_mask, mY_fltr_mask, dem_mask_ds.GetGeoTransform())
    pX_fltr_mask = np.atleast_1d(pX_fltr_mask)
    pY_fltr_mask = np.atleast_1d(pY_fltr_mask)

    dz = z_fltr_mask - samp[samp_idx,0]

    if True:
        print "Creating plot of %i output points" % x_fltr.shape[0]
        fig_kw = {'figsize':(10,7.5)}
        fig, (ax1, ax2, ax3, ax4) = plt.subplots(1, 4, sharex=True, sharey=True, **fig_kw)

        #Plot DEM color shaded relief
        hs_ma = geolib.gdaldem_wrapper(dem_fn)
        hs_clim = malib.calcperc(hs_ma, perc=(0.5, 99.5))
        dem_clim = malib.calcperc(dem_ma)
        ax1.imshow(hs_ma, cmap='gray', clim=hs_clim)
        im1 = ax1.imshow(dem_ma, cmap=cpt_rainbow, clim=dem_clim, alpha=0.5)
        cbar = pltlib.add_cbar(ax1, im1, label='DEM Elev. (m WGS84)')
       
        #Plot all color points over shaded relief
        im2 = ax2.imshow(hs_ma, cmap='gray', clim=hs_clim, alpha=0.5)
        #Plot all points in black
        sc2 = ax2.scatter(pX_fltr, pY_fltr, s=0.5, c='k', edgecolors='none')
        #Plot valid in color
        c = z_fltr_mask 
        sc2 = ax2.scatter(pX_fltr_mask, pY_fltr_mask, s=0.5, c=c, cmap=cpt_rainbow, vmin=dem_clim[0], vmax=dem_clim[1], edgecolors='none')
        cbar = pltlib.add_cbar(ax2, sc2, label='Pt Elev. (m WGS84)')

        #Plot time
        c = glas_pts_fltr[:,tcol]
        c_decyear = timelib.np_dt2decyear(timelib.np_o2dt(c))
        c = c_decyear
        #vmin = c.min()
        #vmax = c.max()
        vmin = 2003.14085699
        vmax = 2009.77587047
        #vmin = 20030220
        #vmax = 20091011
        im3 = ax3.imshow(hs_ma, cmap='gray', clim=hs_clim, alpha=0.5)
        sc3 = ax3.scatter(pX_fltr, pY_fltr, s=1, c=c, vmin=vmin, vmax=vmax, edgecolors='none')
        #cbar = pltlib.add_cbar(ax3, sc3, label='Pt Year', cbar_kwargs={'format':'%0.2f'})
        cbar = pltlib.add_cbar(ax3, sc3, label='Pt Year')

        #Plot dz
        c = dz
        vmin, vmax = malib.calcperc(c, perc=(5, 95))
        absmax = np.max(np.abs([vmin, vmax]))
        vmin = -absmax
        vmax = absmax
        im4 = ax4.imshow(hs_ma, cmap='gray', clim=hs_clim, alpha=0.5)
        sc4 = ax4.scatter(pX_fltr_mask, pY_fltr_mask, s=2, c=c, cmap='RdYlBu', vmin=vmin, vmax=vmax, edgecolors='none')
        cbar = pltlib.add_cbar(ax4, sc4, label='GCP - DEM (m)')

        for ax in (ax1, ax2, ax3, ax4):
            ax.xaxis.set_visible(False)
            ax.yaxis.set_visible(False)
            ax.set_aspect('equal', 'box-forced')

        title='%s \n %i valid points (%i initial)' % (os.path.splitext(os.path.split(dem_fn)[1])[0], pX_fltr_mask.shape[0], pX_fltr.shape[0])
        fig.suptitle(title)
        fig.tight_layout()
        #This adjusts subplots to fit suptitle
        plt.subplots_adjust(top=0.92)
        fig_fn = os.path.splitext(out_csv_fn)[0]+'.png'
        print "Saving figure: %s" % fig_fn
        plt.savefig(fig_fn, dpi=300, bbox_inches='tight', pad_inches=0)
        plt.close(fig)



#! /usr/bin/env python

#David Shean
#dshean@gmail.com

#Little utility for volume change analysis from dz rasters
#Input should already be clipped appropriately (e.g. masked except over glaciers)

import sys
import os
import datetime
import argparse

import numpy as np

from pygeotools.lib import iolib
from pygeotools.lib import malib
from pygeotools.lib import geolib
from pygeotools.lib import timelib

def getparser():
    parser = argparse.ArgumentParser(description="Compute volume/mass change stats from DEM difference")
    parser.add_argument('fn', type=str, help='Elevation difference filename (dz.tif)')
    parser.add_argument('-rho', type=float, default=0.917, help='Density for mass change calculation (default: %(default)s)')
    return parser

def main():
    parser = getparser()
    args = parser.parse_args()

    fn = args.fn

    #This is mean density for N Cascades snow
    #rho = 0.5
    #Density of pure ice 
    rho = args.rho
    #If number is in kg/m^3 rather than g/cc
    if rho > 10.:
        rho /= 1000.

    #Clip negative values to 0
    filt = False 

    src_ds = iolib.fn_getds(fn)
    res = geolib.get_res(src_ds, square=True)[0]
    bma = iolib.ds_getma(src_ds)

    #Attempt to extract t1 and t2 from input filename
    ts = timelib.fn_getdatetime_list(fn)
    #Hardcode timestamps
    #ts = [datetime.datetime(2013,9,10), datetime.datetime(2014,5,14)]

    dt_yr = None
    if len(ts) == 2:
        dt = ts[1] - ts[0]
        year = datetime.timedelta(days=365.25)
        dt_yr = dt.total_seconds()/year.total_seconds()

    #Can add filter here to remove outliers, perc_fltr(0.01, 99.9)
    if filt:
        mask = np.ma.getmaskarray(bma)
        bma[bma < 0] = 0
        bma = np.ma.array(bma, mask=mask)

    #Print out stats
    print('\n')
    stats = malib.print_stats(bma)
    print('\n')

    count = stats[0]
    area = res**2*count
    mean = stats[3]
    med = stats[5]

    s_m3 = np.ma.sum(bma)*res**2 
    s_km3 = s_m3/1E9 
    s_mwe = mean*rho
    s_gt = s_km3*rho
    #s_mm = s_gt/374
    #https://climatesanity.wordpress.com/conversion-factors-for-ice-and-water-mass-and-volume/
    s_mm = s_gt/360

    if dt_yr is not None:
        print("%s to %s: %0.2f yr" % (ts[0], ts[1], dt_yr))
        print("%0.0f m^3 (%0.0f m^3/yr)" % (s_m3, s_m3/dt_yr))
        print("%0.3f km^3 (%0.3f km^3/yr)" % (s_km3, s_km3/dt_yr))
        print("Density: %0.3f g/cc" % rho)
        print("%0.3f GT (%0.3f GT/yr)" % (s_gt, s_gt/dt_yr))
        print("%0.6f mm SLR (%0.6f mm/yr)" % (s_mm, s_mm/dt_yr))
        print("%0.3f m.w.e. (%0.3f m.w.e./yr)" % (s_mwe, s_mwe/dt_yr))
    else:
        print("Area: %0.2f km2" % (area/1E6))
        print("%0.0f m^3" % s_m3)
        print("%0.3f km^3" % s_km3) 
        print("Density: %0.3f g/cc" % rho)
        print("%0.3f GT" % s_gt)
        print("%0.6f mm SLR" % s_mm)
        print("%0.3f m.w.e." % s_mwe)
    print('\n')

if __name__ == "__main__":
    main()



#! /usr/bin/env python

"""
Create anomaly map time series
"""

import os
import sys
import glob

import numpy as np
import matplotlib.pyplot as plt

from pygeotools.lib import iolib, timelib, malib, geolib
from imview.lib import pltlib

def makefig(dem, hs, anomaly, ds, title=None):
    f,axa = plt.subplots(2, figsize=(8,5))
    hs_clim = (1, 255)
    hs_im = axa[0].imshow(hs, vmin=hs_clim[0], vmax=hs_clim[1], cmap='gray')
    #dem_clim = (1600, 2100)
    dem_im = axa[0].imshow(dem, vmin=dem_clim[0], vmax=dem_clim[1], cmap='cpt_rainbow', alpha=0.5)
    anomaly_clim = (-15, 15)
    anomaly_im = axa[1].imshow(anomaly, vmin=anomaly_clim[0], vmax=anomaly_clim[1], cmap='RdBu')
    pltlib.add_cbar(axa[0], dem_im, label='Elevation (m WGS84)')
    pltlib.add_cbar(axa[1], anomaly_im, label='Elevation Anomaly (m)')
    res = 8
    pltlib.add_scalebar(axa[0], res=res)
    plt.tight_layout()

outdir = 'stack_anomaly'
if not os.path.exists(outdir):
    os.makedirs(outdir)

#dem_fn_list = glob.glob('*8m_trans_warp.tif')
#stack = malib.DEMStack(dem_fn_list, med=True)
#dem_ref_fn = 'rainier_allgood_mos-tile-0_warp.tif'
#dem_ref = iolib.fn_getma(dem_ref_fn)

stack_fn = sys.argv[1]
#stack = malib.DEMStack(stack_fn=stack_fn, med=True)
#dem_ref = stack.stack_med
stack = malib.DEMStack(stack_fn=stack_fn)
dem_ref = stack.stack_mean
dem_ds = stack.get_ds()
dem_clim = malib.calcperc(dem_ref, (2,98))
dem_fn_list = stack.fn_list 
anomaly_stack = stack.ma_stack - dem_ref
anomaly_clim = np.max(np.abs(malib.calcperc(anomaly_stack, (1,99))))
anomaly_clim = (-anomaly_clim, anomaly_clim)

#for dem_fn in [dem_ref_fn]+dem_fn_list:
for n, dem_fn in enumerate(dem_fn_list):
    print('%i of %i: %s' % (n+1, len(dem_fn_list), dem_fn))
    #print(dem_fn)
    #dem_ds = iolib.fn_getds(dem_fn)
    #dem = iolib.ds_getma(dem_ds)
    dem_fn = stack.fn_list[n]
    #title = dem_fn
    title = None
    dem = stack.ma_stack[n]
    anomaly = anomaly_stack[n]
    #dem_clim = malib.calcperc(stack.ma_stack, (2,98))
    #dem_hs_fn = os.path.splitext(dem_fn)[0]+'_hs_az315.tif'
    #dem_hs = iolib.fn_getma(dem_hs_fn)
    #dem_hs = geolib.gdaldem_mem_ma(dem, dem_ds, returnma=True)
    dem_hs = geolib.gdaldem_mem_ds(dem_ds, returnma=True)
    #dt = timelib.fn_getdatetime(dem_fn)
    dt = stack.date_list[n]
    if dt is not None:
        title = dt.strftime('%Y-%m-%d')
    #f = makefig(dem, dem_hs, anomaly, ds=dem_ds, title=title)
    f,ax = plt.subplots()
    im = ax.imshow(anomaly, clim=anomaly_clim, cmap='RdBu')
    pltlib.add_cbar(ax, im, label='Elevation anomaly (m)')
    pltlib.add_scalebar(ax, res=stack.res, location='lower left')
    pltlib.hide_ticks(ax)
    ax.set_facecolor('k')
    if title is not None:
        ax.set_title(title)
    out_fn = os.path.join(outdir, os.path.splitext(os.path.split(dem_fn)[-1])[0]+'_anomaly.png')
    f.savefig(out_fn, bbox_inches='tight', dpi=150)




#! /usr/bin/env python

#David Shean
#dshean@gmail.com

import os
import sys
import argparse

import numpy as np
from pygeotools.lib import iolib

#This takes in the sampled output from pc_align and computes statistics
#Compute and print robust statistics for before or after samples (csv or tif) from pc_align_wrapper.sh

def getparser():
    parser = argparse.ArgumentParser(description="Compute robust stats from dz or pc_align sample.csv")
    parser.add_argument('fn', type=str, help='Input filename')
    parser.add_argument('-outdir', default=None, help='Output directory')
    parser.add_argument('-col', type=int, default=4, help='Column containing values for stats')
    return parser

def main():
    parser = getparser()
    args = parser.parse_args()

    fn = args.fn
    ext = os.path.splitext(fn)[1]

    col = args.col

    #This expects the sample.csv, not the errors.csv
    if 'csv' in ext:
        #a = np.loadtxt(fn, delimiter=',', skiprows=1)
        #This will more gracefully handle hader
        a = iolib.readcsv(fn)
        #Signed difference values are in column 5
        dz_m = a[:,col]
        dz_m = dz_m[~np.isnan(dz_m)]
    #If pc_align was run with reference grid, then load the dz raster
    elif 'tif' in ext: 
        a = iolib.fn_getma(fn)
        dz_m = a.compressed()
    else:
        sys.exit('Unsupported input type')

    dz_m_abs = np.abs(dz_m)

    #Extract fn date
    #d = f[0:13]
    #print("Date: %s" % d)

    #print("Filename: %s" % f)

    print("Count: %i" % (dz_m.shape[0] - 1))

    rmse = np.sqrt(np.sum(dz_m**2)/dz_m.size)
    print("RMSE: %0.3f" % rmse)

    mean = np.mean(dz_m)
    print("Mean Error: %0.3f" % mean) 

    std = np.std(dz_m)
    print("Standard Deviation: %0.3f" % std) 

    #thresh = 3 * std
    med = np.median(dz_m)
    print("Median Error: %0.3f" % med) 

    p16, p84 = np.percentile(dz_m, (15.9, 84.2))
    spread = p84 - p16
    print("16th Percentile: %0.3f" % p16)
    print("84th Percentile: %0.3f" % p84)
    print("Spread: %0.3f" % spread)

    absmed = np.median(dz_m_abs)
    print("Absolute Median Error: %0.3f" % absmed) 

    mad = np.median(np.abs(dz_m - med))
    #print("MAD: %0.3f" % mad)
    nmad = 1.4826 * mad
    print("NMAD: %0.3f" % nmad)

    p68, p95 = np.percentile(dz_m_abs, (68.3, 95.0))
    print("68.3th Percentile: %0.3f" % p68)
    print("95th Percentile: %0.3f" % p95)

    #se = np.std(dz_m) / np.sqrt(dz_m.size)
    #print "Standard Error: %0.3f" % se 

if __name__ == "__main__":
    main()



#! /usr/bin/env python

"""
Library of functions that can be used for co-registration of raster data

For many situations, ASP pc_align ICP co-registration is superior to these approaches. See pc_align_wrapper.sh
"""
import sys
import numpy as np
import matplotlib.pyplot as plt

from pygeotools.lib import malib, iolib

def apply_xy_shift(ds, dx, dy, createcopy=True):
    """
    Apply horizontal shift to GDAL dataset GeoTransform
    
    Returns:
    GDAL Dataset copy with updated GeoTransform
    """
    print("X shift: ", dx)
    print("Y shift: ", dy)
   
    #Update geotransform
    gt_orig = ds.GetGeoTransform()
    gt_shift = np.copy(gt_orig)
    gt_shift[0] += dx 
    gt_shift[3] += dy

    print("Original geotransform:", gt_orig)
    print("Updated geotransform:", gt_shift)

    #Update ds Geotransform
    if createcopy:
        ds_align = iolib.mem_drv.CreateCopy('', ds, 0)
    else:
        #Update in place, assume ds is opened as GA_Update
        ds_align = ds
    ds_align.SetGeoTransform(gt_shift)
    return ds_align

def apply_z_shift(ds, dz, createcopy=True):
    if isinstance(dz, np.ndarray):
        print("Z shift offset array mean: ", dz.mean())
    else:
        print("Z shift offset: ", dz) 
    if createcopy:
        ds_shift = iolib.mem_drv.CreateCopy('', ds, 0)
    else:
        ds_shift = ds
    b = ds_shift.GetRasterBand(1)
    a = iolib.b_getma(b)
    a += dz
    b.WriteArray(a.filled())
    return ds_shift

#Function for fitting Nuth and Kaab (2011)
def nuth_func(x, a, b, c):
    y = a * np.cos(np.deg2rad(b-x)) + c
    #Can use Phasor addition, but need to change conversion to offset dx and dy
    #https://stackoverflow.com/questions/12397412/i-know-scipy-curve-fit-can-do-better?rq=1
    #y = a * np.cos(np.deg2rad(x)) + b * np.sin(np.deg2rad(x)) + c
    return y

def compute_offset_sad(dem1, dem2, pad=(9,9), plot=False):
    """Compute subpixel horizontal offset between input rasters using sum of absolute differences (SAD) method
    """
    #This defines the search window size
    #Use half-pixel stride?
    #Note: stride is not properly implemented 
    #stride = 1
    #ref = dem1[::stride,::stride]
    #kernel = dem2[pad[0]:-pad[0]:stride, pad[1]:-pad[1]:stride]
    kernel = dem2[pad[0]:-pad[0], pad[1]:-pad[1]]
    #Want to pad evenly on both sides, so add +1 here
    m = np.zeros((pad[0]*2+1, pad[1]*2+1))
   
    #Find integer pixel offset
    i = j = 0
    for i in range(m.shape[0]):
        print(i)
        for j in range(m.shape[1]):
            print(j)
            ref = dem1[i:i+kernel.shape[0], j:j+kernel.shape[1]]
            diff = ref - kernel
            
            #Remove outliers beyond IQR
            diff_iqr = malib.calcperc(diff, (25,75))
            diff = np.ma.masked_outside(diff, *diff_iqr)
            """ 
            diff_med = np.ma.median(diff)
            diff_mad = malib.mad(diff)
            diff_madr = (diff_med - mad, diff_med + mad)
            diff = np.ma.masked_outside(diff, diff_madr)     
            """
            #Masked areas will decrease sum! Normalize by count of valid pixels
            m[i,j] = np.ma.abs(diff).sum()/diff.count()
    
    #Note, we're dealing with min SAD here, so want to provide -m for sub-pixel refinement 
    m = -m  

    int_argmax = np.array(np.unravel_index(m.argmax(), m.shape))
    int_offset = int_argmax - pad
    
    sp_argmax = np.array(find_subpixel_peak_position(m, 'parabolic'))
    sp_offset = sp_argmax - pad

    if plot:
        plt.figure()
        plt.title('Sum of Absolute Differences')
        plt.imshow(m)
        plt.scatter(*sp_argmax[::-1])
        #plt.show()

    return m, int_offset, sp_offset

#This is a decent full-image normalized cross-correlation routine with sub-pixel refinement
def compute_offset_ncc(dem1, dem2, pad=(9,9), prefilter=False, plot=False): 
    """Compute horizontal offset between input rasters using normalized cross-correlation (NCC) method
    """

    #Apply edge detection filter up front - improves results when input DEMs are same resolution
    if prefilter:
        print("Applying LoG edge-detection filter to DEMs")
        sigma = 1
        import scipy.ndimage
        #Note, ndimage alone propagates Nans and greatly reduces valid data area
        #Use the malib.nanfill wrapper to avoid this
        dem1 = malib.nanfill(dem1, scipy.ndimage.filters.gaussian_laplace, sigma) 
        dem2 = malib.nanfill(dem2, scipy.ndimage.filters.gaussian_laplace, sigma) 

    import scipy.signal
    #Compute max offset given dem spatial resolution
    #Should implement arbirary x and y search space
    #xsearch = (20, 41)
    #ysearch = (-10, 1)
    stride = 1
    ref = dem1[::stride,::stride]
    kernel = dem2[pad[0]:-pad[1]:stride, pad[0]:-pad[1]:stride]
    #kernel = dem2[-ysearch[0]:-ysearch[1]:stride, xsearch[0]:-xsearch[1]:stride]

    #Normalize
    ref = (ref - ref.mean()) / ref.std()
    kernel = (kernel - kernel.mean()) / kernel.std()

    #Consider using astropy.convolve here instead of scipy.correlate?

    print("Adding random noise to masked regions")
    #Generate random noise to fill gaps before correlation in frequency domain
    #Normal distribution N(mean, std^2)
    #ref_noise = ref.mask * ref.std() * np.random.rand(*ref.shape) + ref.mean()
    #kernel_noise = kernel.mask * kernel.std() * np.random.rand(*kernel.shape) + kernel.mean()
    #This provides noise in proper range, but noise propagates to m, peak is in different locations!
    #ref_noise = ref.mask * (ref.min() + ref.ptp() * np.random.rand(*ref.shape))
    #kernel_noise = kernel.mask * (kernel.min() + kernel.ptp() * np.random.rand(*kernel.shape))

    #This provides a proper normal distribution with mean=0 and std=1
    ref_noise = ref.mask * (np.random.randn(*ref.shape))
    kernel_noise = kernel.mask * (np.random.randn(*kernel.shape))
    #Add the noise
    ref = ref.filled(0) + ref_noise
    kernel = kernel.filled(0) + kernel_noise

    print("Running 2D correlation with search window (x,y): %i, %i" % (pad[1], pad[0]))
    m = scipy.signal.correlate2d(ref, kernel, 'valid')
    #This has memory issues, but ndimage filters can handle nan
    #m = scipy.ndimage.filters.correlate(ref, kernel)
   
    print("Computing sub-pixel peak")
    int_argmax = np.array(np.unravel_index(m.argmax(), m.shape))
    int_offset = int_argmax*stride - pad
    #int_offset = int_argmax*stride + np.array([ysearch[0], xsearch[0]]) 

    print(m.argmax())
    print(m.shape)
    print(int_argmax)
    print(int_offset)

    #Find sub-pixel peak
    sp_argmax = np.array(find_subpixel_peak_position(m, 'parabolic'))
    #May need to split this into integer and decimal components, multipy stride*int and add decimal
    #sp_offset = int_offset + (sp_argmax - int_argmax)
    sp_offset = sp_argmax - pad
    #sp_offset = sp_argmax + np.array([ysearch[0], xsearch[0]]) 

    print(sp_argmax)
    print(sp_offset)

    if plot: 
        fig, ax = plt.subplots()
        ax.set_title('NCC offset, parabolic SPR')
        ax.imshow(m)
        #plt.scatter(*int_argmax[::-1])
        ax.scatter(*sp_argmax[::-1])
    else:
        fig = None

    return m, int_offset, sp_offset, fig

#This is the Nuth and Kaab (2011) method
def compute_offset_nuth(dh, slope, aspect, min_count=100, remove_outliers=True, plot=True):
    """Compute horizontal offset between input rasters using Nuth and Kaab [2011] (nuth) method
    """
    import scipy.optimize as optimization

    if dh.count() < min_count:
        sys.exit("Not enough dh samples")
    if slope.count() < min_count:
        sys.exit("Not enough slope/aspect samples")

    #mean_dh = dh.mean()
    #mean_slope = slope.mean()
    #c_seed = (mean_dh/np.tan(np.deg2rad(mean_slope))) 
    med_dh = malib.fast_median(dh)
    med_slope = malib.fast_median(slope)
    c_seed = (med_dh/np.tan(np.deg2rad(med_slope))) 

    x0 = np.array([0.0, 0.0, c_seed])
  
    print("Computing common mask")
    common_mask = ~(malib.common_mask([dh, aspect, slope]))

    #Prepare x and y data
    xdata = aspect[common_mask].data
    ydata = (dh[common_mask]/np.tan(np.deg2rad(slope[common_mask]))).data

    print("Initial sample count:")
    print(ydata.size)

    if remove_outliers:
        print("Removing outliers")
        #print("Absolute dz filter: %0.2f" % max_dz)
        #diff = np.ma.masked_greater(diff, max_dz)
        #print(diff.count())

        #Outlier dz filter
        f = 3
        sigma, u = (ydata.std(), ydata.mean())
        #sigma, u = malib.mad(ydata, return_med=True)
        rmin = u - f*sigma
        rmax = u + f*sigma
        print("3-sigma filter: %0.2f - %0.2f" % (rmin, rmax))
        idx = (ydata >= rmin) & (ydata <= rmax)
        xdata = xdata[idx]
        ydata = ydata[idx]
        print(ydata.size)

    #Generate synthetic data to test curve_fit
    #xdata = np.arange(0,360,0.01)
    #ydata = f(xdata, 20.0, 130.0, -3.0) + 20*np.random.normal(size=len(xdata))
    
    #Limit sample size
    #n = 10000
    #idx = random.sample(range(xdata.size), n)
    #xdata = xdata[idx]
    #ydata = ydata[idx]

    #Compute robust statistics for 1-degree bins
    nbins = 360
    bin_range = (0., 360.)
    bin_width = 1.0
    bin_count, bin_edges, bin_centers = malib.bin_stats(xdata, ydata, stat='count', nbins=nbins, bin_range=bin_range)
    bin_med, bin_edges, bin_centers = malib.bin_stats(xdata, ydata, stat='median', nbins=nbins, bin_range=bin_range)
    #Needed to estimate sigma for weighted lsq
    #bin_mad, bin_edges, bin_centers = malib.bin_stats(xdata, ydata, stat=malib.mad, nbins=nbins, bin_range=bin_range)
    #Started implementing this for more generic binning, needs testing
    #bin_count, x_bin_edges, y_bin_edges = malib.get_2dhist(xdata, ydata, \
    #        xlim=bin_range, nbins=(nbins, nbins), stat='count')

    """
    #Mask bins in grid directions, can potentially contain biased stats
    #Especially true for SGM algorithm
    #badbins = [0, 90, 180, 270, 360]
    badbins = [0, 45, 90, 135, 180, 225, 270, 315, 360]
    bin_stat = np.ma.masked_where(np.around(bin_edges[:-1]) % 45 == 0, bin_stat)
    bin_edges = np.ma.masked_where(np.around(bin_edges[:-1]) % 45 == 0, bin_edges)
    """

    #Remove any bins with only a few points
    min_bin_sample_count = 9
    idx = (bin_count.filled(0) >= min_bin_sample_count) 
    bin_count = bin_count[idx].data
    bin_med = bin_med[idx].data
    #bin_mad = bin_mad[idx].data
    bin_centers = bin_centers[idx]

    fit = None
    fit_fig = None

    #Want a good distribution of bins, at least 1/4 to 1/2 of sinusoid, to ensure good fit
    #Need at least 3 valid bins to fit 3 parameters in nuth_func
    #min_bin_count = 3
    min_bin_count = 90 
    
    #Not going to help if we have a step function between two plateaus, but better than nothing
    #Calculate bin aspect spread
    bin_ptp = np.cos(np.radians(bin_centers)).ptp()
    min_bin_ptp = 1.0 

    #Should iterate here, if not enough bins, increase bin width
    if len(bin_med) >= min_bin_count and bin_ptp >= min_bin_ptp:

        print("Computing fit")
        #Unweighted fit
        fit = optimization.curve_fit(nuth_func, bin_centers, bin_med, x0)[0]

        #Weight by observed spread in each bin 
        #sigma = bin_mad
        #fit = optimization.curve_fit(nuth_func, bin_centers, bin_med, x0, sigma, absolute_sigma=True)[0]

        #Weight by bin count
        #sigma = bin_count.max()/bin_count
        #fit = optimization.curve_fit(nuth_func, bin_centers, bin_med, x0, sigma, absolute_sigma=False)[0]

        print(fit)

        if plot:
            print("Generating Nuth and Kaab plot")
            bin_idx = np.digitize(xdata, bin_edges)
            output = []
            for i in np.arange(1, len(bin_edges)):
                output.append(ydata[bin_idx==i])
            #flierprops={'marker':'.'}
            lw = 0.25
            whiskerprops={'linewidth':lw}
            capprops={'linewidth':lw}
            boxprops={'facecolor':'k', 'linewidth':0}
            medianprops={'marker':'o', 'ms':1, 'color':'r'}
            fit_fig, ax = plt.subplots(figsize=(6,6))
            #widths = (bin_width/2.0)
            widths = 2.5*(bin_count/bin_count.max())
            #widths = bin_count/np.percentile(bin_count, 50)
            #Stride
            s=3
            #This is inefficient, but we have list of arrays with different length, need to filter
            #Reduntant with earlier filter, should refactor
            bp = ax.boxplot(np.array(output)[idx][::s], positions=bin_centers[::s], widths=widths[::s], showfliers=False, \
                    patch_artist=True, boxprops=boxprops, whiskerprops=whiskerprops, capprops=capprops, \
                    medianprops=medianprops)
            bin_ticks = [0, 45, 90, 135, 180, 225, 270, 315, 360]
            ax.set_xticks(bin_ticks)
            ax.set_xticklabels(bin_ticks)
            """
            #Can pull out medians from boxplot
            #We are computing multiple times, inefficient
            bp_bin_med = []
            for medline in bp['medians']:
                bp_bin_med.append(medline.get_ydata()[0])
            """

            #Plot the fit
            f_a = nuth_func(bin_centers, fit[0], fit[1], fit[2])
            nuth_func_str = r'$y=%0.2f*cos(%0.2f-x)+%0.2f$' % tuple(fit)
            ax.plot(bin_centers, f_a, 'b', label=nuth_func_str)

            ax.set_xlabel('Aspect (deg)')
            ax.set_ylabel('dh/tan(slope) (m)')
            ax.axhline(color='gray', linewidth=0.5)

            ax.set_xlim(*bin_range)
            ylim = ax.get_ylim()
            abs_ylim = np.max(np.abs(ylim))
            #abs_ylim = np.max(np.abs([ydata.min(), ydata.max()]))
            #pad = 0.2 * abs_ylim 
            pad = 0
            ylim = (-abs_ylim - pad, abs_ylim + pad)
            minylim = (-10,10)
            if ylim[0] > minylim[0]:
                ylim = minylim
            ax.set_ylim(*ylim)
            ax.legend(prop={'size':8})

    return fit, fit_fig

#Attempt to fit polynomial functions to along-track and cross-track signals
#See demtools for existing code
def fit_at_ct():
    #Derive from image corners in projected array
    #Use known orbintal inclinations, project wgs geometry into srs
    img1_inc
    img2_inc
    #Rotate
    #Stats for rows, cols
    #Fit

#Function copied from from openPIV pyprocess
def find_first_peak(corr):
    """
    Find row and column indices of the first correlation peak.
    
    Parameters
    ----------
    corr : np.ndarray
        the correlation map
        
    Returns
    -------
    i : int
        the row index of the correlation peak
        
    j : int
        the column index of the correlation peak    
    
    corr_max1 : int
        the value of the correlation peak
    
    Original code from openPIV pyprocess

    """    
    ind = corr.argmax()
    s = corr.shape[1] 
    
    i = ind // s 
    j = ind %  s
    
    return i, j, corr.max()

#Function copied from from openPIV pyprocess
def find_subpixel_peak_position(corr, subpixel_method='gaussian'):
    """
    Find subpixel approximation of the correlation peak.
    
    This function returns a subpixels approximation of the correlation
    peak by using one of the several methods available. If requested, 
    the function also returns the signal to noise ratio level evaluated 
    from the correlation map.
    
    Parameters
    ----------
    corr : np.ndarray
        the correlation map.
        
    subpixel_method : string
         one of the following methods to estimate subpixel location of the peak: 
         'centroid' [replaces default if correlation map is negative], 
         'gaussian' [default if correlation map is positive], 
         'parabolic'.
         
    Returns
    -------
    subp_peak_position : two elements tuple
        the fractional row and column indices for the sub-pixel
        approximation of the correlation peak.

    Original code from openPIV pyprocess

    """
    # initialization
    default_peak_position = (corr.shape[0]/2,corr.shape[1]/2)

    # the peak locations
    peak1_i, peak1_j, dummy = find_first_peak(corr)
    
    try:
        # the peak and its neighbours: left, right, down, up
        c = corr[peak1_i, peak1_j]
        cl = corr[peak1_i-1, peak1_j]
        cr = corr[peak1_i+1, peak1_j]
        cd = corr[peak1_i, peak1_j-1] 
        cu = corr[peak1_i, peak1_j+1]
        
        # gaussian fit
        if np.any(np.array([c,cl,cr,cd,cu]) < 0) and subpixel_method == 'gaussian':
            subpixel_method = 'centroid'
        
        try: 
            if subpixel_method == 'centroid':
                subp_peak_position = (((peak1_i-1)*cl+peak1_i*c+(peak1_i+1)*cr)/(cl+c+cr),
                                    ((peak1_j-1)*cd+peak1_j*c+(peak1_j+1)*cu)/(cd+c+cu))
        
            elif subpixel_method == 'gaussian':
                subp_peak_position = (peak1_i + ((np.log(cl)-np.log(cr))/(2*np.log(cl) - 4*np.log(c) + 2*np.log(cr))),
                                    peak1_j + ((np.log(cd)-np.log(cu))/( 2*np.log(cd) - 4*np.log(c) + 2*np.log(cu)))) 
        
            elif subpixel_method == 'parabolic':
                subp_peak_position = (peak1_i +  (cl-cr)/(2*cl-4*c+2*cr),
                                        peak1_j +  (cd-cu)/(2*cd-4*c+2*cu)) 
    
        except: 
            subp_peak_position = default_peak_position
            
    except IndexError:
            subp_peak_position = default_peak_position
            
    return subp_peak_position[0], subp_peak_position[1]

